-- Convert logic synchronous to the pp_dck to logic synchrounous to clk

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity epp_asynch is
	port (clk: in STD_LOGIC; 
	  arst: in STD_LOGIC; -- asynchronous global reset
    rnw: out std_logic; -- read not write
    adr_not_data: out std_logic; -- transfer of address or data (valid only for rnw='0')
    din: in std_logic_vector(1 downto 0); -- input data
    dout: out std_logic_vector(3 downto 0); -- output data
    stb: out std_logic; -- ready to transfer data
    ack: in std_logic; -- external device is ready to accept/send a new word
    timeout_rst: out std_logic; -- timeout synch. reset - timeout occures when drd indicates a wait state and pp_dck rises
    -- external pins
    pp_dck: in std_logic; -- external PP clock
    pp_dwr: in std_logic; -- external PP data in (write)
    pp_drd: out std_logic); -- external PP data out (read)
end epp_asynch;

architecture arch of epp_asynch is
-------------------------------------------------------------------------------------
-----  Synchronous Logic
  signal pp_dck_q, pp_dck_qq, pp_dwr_q, ce: std_logic;
  signal count: std_logic_vector(2 downto 0); -- state of the machine
  signal dwr: std_logic_vector(3 downto 0); -- data out (write)
  signal drd: std_logic_vector(1 downto 0); -- data in (read)
  signal rnw1, adr_not_data1, dwr_valid, drd_valid, timeout_rst1: std_logic;
  signal adr_not_data_q: std_logic; -- internal states 
  
begin

 --convert asynchronous logic to synchronous
  asynch: process(arst, clk) begin
    if arst='1' then
      pp_dck_q<= '0'; pp_dck_qq<= '0'; pp_dwr_q<= '0';
    elsif clk'event and clk='1' then
      pp_dck_q<= pp_dck;
      pp_dck_qq<= pp_dck_q;
      pp_dwr_q<= pp_dwr;
    end if;
  end process;
  
  ce<= pp_dck_q and not pp_dck_qq;
  
  cnt: process(arst, clk) begin
    if arst='1' then
      count<= (others=>'0'); timeout_rst1<= '0'; rnw1<= '0'; adr_not_data1<= '0';
      dwr_valid<= '0'; drd_valid<= '0';  drd<=(others=>'0');  dwr<=(others=>'0');
    elsif clk'event and clk='1' then
      if count=0 and ce='1' then
        adr_not_data_q<= pp_dwr_q;
      end if;      
      if count=1 and ce='1' then
        rnw1<= pp_dwr_q;
      end if;
      
      if dwr_valid='0' then -- update adr_not_data1 only when write buffer is not full
        adr_not_data1<= adr_not_data_q;
      end if;

      
      if count>1 and ce='1' then
        dwr<= pp_dwr_q & dwr(3 downto 1);  
      end if;
      
      if count=5 and ce='1' then
        dwr_valid<= '1';
      elsif ack='1' or timeout_rst1='1' then
        dwr_valid<= '0';
      end if;
      
      if ack='1' then
        drd<= din;
      end if;
      
      if ack='1' and rnw1='1' then
        drd_valid<= '1';
      elsif count= 4 and ce='1' then
        drd_valid<= '0';
      end if;
      
      if ce='1' then
        if count>=5 or (count=4 and rnw1='1') then
          count<= (others=>'0');
        else
          count<= count + 1;
        end if;
      end if;
      
      if count=2 and ce='1' and ((rnw1='1' and drd_valid='0') or dwr_valid='1') then
        timeout_rst1<= '1';
      else
        timeout_rst1<= '0';
      end if;
      
      if count<2 then
        pp_drd<= '0'; -- default wait state (read / write direction is not known)
      elsif count=2 then
        if rnw1='0' then -- writing
          pp_drd<= not dwr_valid;
        else -- reading
          if ((pp_dck='1' and pp_dck_q='0') or ce='1') then -- swithing to state 3 - improve read latency
            pp_drd<= drd(0);
          else -- wait state?
            pp_drd<= drd_valid;
          end if;
        end if;
      else -- count>=3 
        if rnw1='0' then --writing
          pp_drd<= '-';
        elsif ((pp_dck='1' and pp_dck_q='0') or ce='1') then -- going to state 4
          pp_drd<= drd(1);
        end if;
      end if;
        
    end if;  -- clk
  end process;
      
  stb<= '1' when dwr_valid='1' else -- write
    not drd_valid and rnw1; -- read
  
  --pp_drd1<= '0' when dwr_valid='1' else -- write buffer is still not empty so wait
  --        when drd_valid='0' and rnw1='1' else -- reading process - wait signal
  --       drd(0); -- data read (or don't care)         
  
  --pp_drd<= pp_drd1;
  rnw<= not dwr_valid;
  adr_not_data<= adr_not_data1;
  timeout_rst<= timeout_rst1;
  dout<= dwr;
 

 
----------------------------------------------------------------------------------------------------
-- Asynch logic   
--  signal dout1: std_logic_vector(3 downto 0); -- internal dout
--  signal din_q: std_logic_vector(1 downto 0); -- internal din
--  signal rnw1, stb1, stb_rd, adr_not_data1: std_logic; -- internal stb
--  signal count: std_logic_vector(2 downto 0); -- counter
--  signal valid_dout, pp_drd1: std_logic;
--  signal ack_q: std_logic; -- ack signal clocked by clk
--  signal arst_valid_dout, arst_ackq_rd, arst_ackq: std_logic;
--  signal timeout_rstq1, timeout_rstq2: std_logic; -- timeout reset Q outputs for different clock signals 
--  signal pp_dck_q, pp_dwr_q: std_logic;
--  --signal pp_rdy: std_logic; -- FPGA is ready to transmit data (pp_drd='1' for count=2)
--  --signal count2: std_logic; -- =1 when count=2
--  --signal pp_drd_q: std_logic; -- read data
--begin
--process(arst, pp_dck) begin
--  if arst='1' then
--    adr_not_data1<= '0'; rnw1<='0'; timeout_rstq1<= '0';
--    count<= (others=>'0');
--  elsif pp_dck'event and pp_dck='1' then
--    if count=0 then -- adress/data
--      adr_not_data1<= pp_dwr;
--    elsif count=1 then -- read not write
--      rnw1<= pp_dwr;
--    else
--      dout1<= pp_dwr & dout1(3 downto 1);
--    end if;
--
--    if count=2 then -- timeout logic
--      timeout_rstq1<= not pp_drd1;  -- sample the wait state on the third clock cycle
--    elsif timeout_rstq2='1' then -- generate only a short time impuls
--      timeout_rstq1<= '0';
--    end if;
--
--    if (rnw1='0' and count= 5) or (rnw1='1' and count=4) then
--      count<= (others=>'0');
--    else 
--      count<= count + 1;
--    end if;
--  end if; 
--end process;
--dout<= dout1;
--
---- timeout logic the second flip-flop
--process(arst, clk) begin
--  if arst='1' then timeout_rstq2<= '0';
--  elsif clk'event and clk='1' then
--    timeout_rstq2<= timeout_rstq1;
--  end if;
--end process;
--timeout_rst<= timeout_rstq2;
--
---- valid_dout
--process(pp_dck, arst, arst_valid_dout) begin
--  if arst='1' or arst_valid_dout='1' then
--    valid_dout<= '0';
--  elsif pp_dck='1' and pp_dck'event then
--    if rnw1='0' and count=5 then
--      valid_dout<='1';
--    end if;
--  end if;
--end process;
--arst_valid_dout<= valid_dout and ack_q;
--
---- read strobe and data read
--process(pp_dck, arst) begin
--  if arst='1' then 
--    stb_rd<='0'; arst_ackq_rd<= '0';
--  elsif pp_dck='1' and pp_dck'event then
--    if count=1 and pp_dwr='1' then -- read data
--      stb_rd<='1';
--    else
--      stb_rd<= '0';
--    end if;
--    arst_ackq_rd<= stb_rd;
--  end if;
--end process;
--
----- adres not data
--process(clk, arst) begin
--  if arst='1' then adr_not_data<= '1';
--  elsif clk='1' and clk'event then
--    if ack='0' and ack_q='0' and (valid_dout='1' or stb_rd='1') then
--      adr_not_data<= adr_not_data1;
--    end if;
--  end if;
--end process;
--
--process(clk, arst, arst_ackq) begin
--  if arst='1' or arst_ackq='1' then
--    stb1<= '0'; ack_q<='0';
--  elsif clk='1' and clk'event then
--    -- stb
--    if ack='0' and ack_q='0' and (valid_dout='1' or stb_rd='1') then
--      stb1<= '1';
--    else
--      stb1<='0';
--    end if;
--    -- ack_q
--    if ack='1' then
--      ack_q<='1';
--      din_q<= din;
--    elsif stb_rd='0' then -- read process is finished
--      ack_q<='0';
--    end if;
--    -- pp_rdy
--    if rnw1='1' then -- reading process
--      
--  end if;
--end process;
--arst_ackq<= arst_valid_dout or arst_ackq_rd;
-- 
--pp_drd1<= '0' when valid_dout='1' else -- write buffer is still not empty so wait
--         '1' when rnw1= '0' else -- no wait for writes
--         ack_q when count=2  else -- ready signal for reads
--         din_q(0) when count=3 else -- data read 0
--         din_q(1); -- data read 1 (or don't care)
--pp_drd<= pp_drd1;
--
--stb<= stb1;
--rnw<= valid_dout;
--
--
       


end arch;