-- opb_epp module
-- this module can transfer data from address 0 to FB FFFF to the OPB bus
-- connects OPB bus to PC by Parallel Port (EPP mode) and epp.exe program
-- Generates System Reset OPB_Rst (Address: 0xFF FFF0 - FF FFFF) (write 1 to set, write 0 to clear)
-- Allows to use internal Logic Analyser to analyse OPB bus activity, LA Base Address= FF C000

-- by Erenst Jamro 
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for global set reset signal

entity opb_epp is
  generic (c_opb_dwidth: integer:= 32; -- OPB data bus width
           c_opb_awidth: integer:= 32; -- OPB address bus width
		   c_real_dwidth: integer:= 8; -- real OPB data width (the rest lSBs are always 0s)
           c_use_seqAddr: integer:= 1; -- use seqAddr for OPB bus transfer				   
           c_la_mwidth: integer:= 9; --  internal memory address width (number of samples= 2**c_la_mwidth
		                             -- if <4 do not instantiate logic analyser at all
		   c_la_run_length_coding: integer:= 0; -- include run length codding logic
		   c_la_dwidth: integer:= 8; -- internal logic analyser data width= 8, 16 or 32
		   c_la_dtype: integer:= 0; -- 0- use internal data, 1- use external data, 2- OPB_ABus, 3- OPB_DBus
           c_la_twidth: integer:= 8; -- trigger width
		   c_la_ttype: integer:= 0; -- trigger data type= 0- internal, 1- external, 2- OPB_ABus, 3- OPB_DBus 4- same as la_dbus
		   c_la_ce_dwidth: integer:= 8; -- width of the data clock enable
   		   c_la_ce_dtype: integer:= 3; -- data clock enable= 0- standard connecttion, 1 - external, 2- same as data, 3- same as trigger
		   c_la_ce_ttype: integer:= 0); -- trigger clock enable=  0- always '1', 1- external,
		   -- 2- '1' when OPB_select='1'; 3 - '1' when OPB_xferAck='1', 4 - '1' when OPB_errAck='1' or OPB_xferAck or OPB_retry='1', 5- OPB_rst
	port (OPB_clk: in STD_LOGIC; 
	    OPB_rst: in STD_LOGIC;
		-- OPB interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO c_opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
	    M_request : out STD_LOGIC;
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO c_opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
		OPB_RNW: in STD_LOGIC;
		OPB_ABus: in std_logic_vector(0 to c_opb_awidth-1);
		OPB_select: in std_logic;  -- used only by LA
		OPB_seqAddr: in STD_LOGIC; -- used only by LA
		OPB_busLock: in STD_LOGIC; -- used only by LA
		OPB_toutSup: in STD_LOGIC; -- used only by LA
		-- generate OPB_RST signal
		SYS_rst: out std_logic;
		-- logic analyser data (triger is the same assert data
		la_DBus: in std_logic_vector(0 to c_la_dwidth-1); -- LA data bus
		la_TBus: in std_logic_vector(0 to c_la_twidth-1); -- LA triger bus
		la_ceDBus: in std_logic_vector(0 to c_la_ce_dwidth-1); -- LA clock enable data bus
		la_ceTBus: in std_logic; -- clock enable for la_TBus
		-- external (PIN) signals connected to CPLD and Parallel Port (FPGA is a slave device to CPLD!)
		pp_dck: in STD_LOGIC; -- parallel port (PP) data clock
		pp_dwr: in STD_LOGIC; -- PP data write
		pp_drd: out std_logic -- PP data read
	);
end opb_epp;

architecture opb_epp_arch of opb_epp is

component epp
	port (clk_I: in STD_LOGIC; 
	  arst: in STD_LOGIC; -- asynchronous global reset
		-- WISHBONE interface
		adr_O: buffer std_logic_vector(31 downto 0);
		we_O, cyc_O, stb_O: buffer std_logic;
		ack_I: in std_logic;
		dat_O: out std_logic_vector(7 downto 0);
		dat_I: in std_logic_vector(7 downto 0);
		cti_O: out std_logic_vector(2 downto 0); -- sequencial addressing
    buf_valid: in std_logic; -- write data in the external buffer are alid so do not change the address (only +1 is allowed)
		-- external (PIN) signals connected to CPLD and Parallel Port (FPGA is a slave device to CPLD!)
		pp_dck: in STD_LOGIC; -- Parallel Port Clock
		pp_dwr: in STD_LOGIC; -- PP data write
		pp_drd: out std_logic -- PP data read
	);
  end component;

component ROC -- reset on Configuration
  port (O : out std_logic);
  end component;

component epp_dwidth -- convert 8-bit to 16 or 32 bit data
  generic(dwidth: integer:= 32;
       use_seqAddr: integer:= 1); -- assume that the read or write is sequancial
  port (clk, arst: in std_logic;
        -- common logic
        seqAddr: in std_logic; -- the next read /writes will be sequencial- use always seqAddr when transfering data unless use_seqAddr<=0
        rnw: in std_logic; -- read not write
        buf_valid: out std_logic; -- valid data write buffer - do not change the address and rnw
        -- 8-bit data interface
        adr8: in std_logic_vector(0 to 31);
        stb8: in std_logic;
        ack8: out std_logic;
        din8: in std_logic_vector(0 to 7);
        dout8: out std_logic_vector(0 to 7);
        -- dwidth data interface
        adr: out std_logic_vector(0 to 31);        
        stb: out std_logic;
        ack: in std_logic;
        be: out std_logic_vector(0 to dwidth/8-1);
        din: in std_logic_vector(0 to dwidth-1);
        dout: out std_logic_vector(0 to dwidth-1));
end component;

component opb_master
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width the MSB of OPB_DBus='0'
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal address bus width
    transfer_size: integer:= 4); -- max transfer size when bus is locked (valid only when seqAddr>0)
  port(clk, arst: std_logic; --clock and global asynchronous reset
		-- OPB Master interface
	    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
	    M_BE : out STD_LOGIC_VECTOR(0 TO opb_dwidth/8 -1); 
	    M_RNW : out STD_LOGIC;
	    M_select : out STD_LOGIC;
	    M_seqAddr : out STD_LOGIC;
	    M_busLock : out STD_LOGIC;
	    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    M_request : out STD_LOGIC;
	    OPB_MGrant: in STD_LOGIC;
	    OPB_pendReq: in STD_LOGIC;
	    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
	    OPB_errAck : in STD_LOGIC;
	    OPB_retry : in STD_LOGIC;
	    OPB_timeout: in STD_LOGIC;
	    OPB_xferAck : in STD_LOGIC;
        -- internal simplified bus
        di: in std_logic_vector(0 to dwidth-1); -- data input
        do: out std_logic_vector(0 to dwidth-1); -- data output
        be: in std_logic_vector(0 to dwidth/8 -1); -- byte enable
        a: in std_logic_vector(0 to awidth-1); -- address bus
        msb_adr: in std_logic_vector(0 to opb_awidth-1):= (others=>'0');
        rnw: in std_logic; -- read not write
	    sel: in std_logic; -- OPB select signal
        seqAddr: in std_logic;
	    ack: out std_logic);
end component;

component log_anal
  generic(data_width: integer:= 16; -- width of the data that are analysed (must be power of 2)
      mem_adr_width: integer:= 9; -- internal memory address width
	  adr_width: integer:= 11; -- adr_I address width
      trig_width: integer:= 8; -- width of the triger logic
	  ce_dwidth: integer:= 1; -- data clock enable trigger-like logic width
	  ced_type: integer:= 1; -- 0 - data clock enable logic is desabled, 1- enable data clock enable logic
	  two_clocks: integer:= 0; -- two seperate clocks for control interface and logic analyser
      use_run_length_coding: integer:= 0);
  -- the following rules must!!! be satysfied for generic:
  -- data_width= 8, 16, 32, 64
  -- adr_width= mem_adr_width + 1 + log2(data_width/8); 4<=mem_adr_width<=16
  -- mem_adr_width<=16
  -- trig_width<= 64
  -- two_clocks= 0 (only wb_clk_I clock is used) 1- f_clk= k*f_wb_clk_I, 2- two asynchronous clocks
  
  -- mem_adr_width optimal selection for 4kb BRAM (Virtex): 
  --  mem_adr_width<= 9 then number of BRAM block used= data_width/8
    
  port (arst: in std_logic; -- global asynchronous set reset signal (mainly for simulation purposes)
	  -- interface for logic analyser 
	  clk: in std_logic; -- seperate clock for logic analyzer (the same as wb_clk_I when two_clocks= 0)
	  data: in std_logic_vector(data_width-1 downto 0); -- data that are analysied
	  ce_data: in std_logic_vector(ce_dwidth-1 downto 0); -- clock enable -- should be used if data are recorded e.g. every second la_clk
      trig: in std_logic_vector(trig_width-1 downto 0); -- triger bus (can be different that data bus
	  ce_trig: in std_logic; -- clock enable for triger bus
	  -- control WISHBOBE slave interface - interface for setting logic analyser options and transfering analysed data to computer
      wb_clk_I: in STD_LOGIC; -- clock (common for every logic every logic analyser signals) - common for loagic analyser and WISHBONE control interface
	  wb_adr_I: in std_logic_vector(adr_width-1 downto 0); -- address bus (one bit wider than mem_adr_width)
	  wb_dat_I: in std_logic_vector(7 downto 0);
	  wb_dat_O: out std_logic_vector(7 downto 0);
	  wb_stb_I, wb_we_I: in std_logic; 
	  -- wb_cyc_I: in std_logic; -- signal is ignored
	  -- wb_rst_I: in std_logic; -- the WISHBONE interface need not be reseted
	  wb_ack_O: buffer std_logic); 
  end component;
	
-- Find LA Address Width
function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
	for i in 0 to 4 loop
	  if 2**i = arg then ret:= i; end if;
	end loop;
    return ret;
  end log2;
	
  signal arst: std_logic; -- asynchronous global reset
  constant la_awidth: integer:= 1 + c_la_mwidth + log2(c_la_dwidth/8);
  constant gnd32: std_logic_vector(31 downto 0):= (others=>'0');
  
  type epp_sel_type is (s_w2o, s_rst, s_la); -- epp which device selects 
    -- (w2o -wishbone2 OPB, s_rst- reset register, s_la<= select Logic analyser)
  signal epp_sel: epp_sel_type;
  
  signal epp_adr: std_logic_vector(31 downto 0); -- epp address
  signal epp_cti: std_logic_vector(2 downto 0); -- sequencial addressing
  signal epp_stb, epp_cyc, epp_ack, epp_we, epp_rnw, epp_buf_valid: std_logic;
  signal w2o_stb, rst_stb, la_stb: std_logic;
  signal w2o_cyc: std_logic;
  signal w2o_ack, rst_ack, la_ack: std_logic;
  signal epp_datW, epp_datR, w2o_dat, rst_dat, la_dat: std_logic_vector(7 downto 0);
  -- wishbone to OPB logic
  signal msel, mack: std_logic;
  signal mdrd, mdwr: std_logic_vector(0 to c_real_dwidth-1);
  signal mbe: std_logic_vector(0 to c_real_dwidth/8 -1);
  signal ma: std_logic_vector(0 to 31);
  -- constant opb_msb_adr: std_logic_vector(0 to c_opb_awidth-1):= (others=>'0');
  
  -- LA signals
  signal la_data: std_logic_vector(c_la_dwidth-1 downto 0);
  signal la_trig: std_logic_vector(c_la_twidth-1 downto 0);
  signal la_ce_trig: std_logic;
  signal la_std_data: std_logic_vector(63 downto 0); -- standard (internal) LA watch signals
  signal la_std_ce: std_logic_vector(0 to 4); -- clock enable standard signals
begin
  epp_sel<= s_rst when epp_adr(31 downto 4)= X"FFFFFFF" else -- address FFFF FFF0 to FFFF FFFF
     s_la when epp_adr(31 downto 16)= X"FFFF" and c_la_mwidth>=4 else  -- address FFFF 0000 to FFFF FFEF
	   s_w2o; 

  epp_ack<= w2o_ack or rst_ack or la_ack;
  epp_datR<= rst_dat when epp_sel=s_rst else
             la_dat when epp_sel=s_la else
	  		 w2o_dat;
  w2o_cyc<= epp_cyc when epp_sel=s_w2o else '0';
  rst_stb<= epp_stb when epp_sel=s_rst else '0';
  la_stb<= epp_stb when epp_sel=s_la else '0';	  
  w2o_stb<= epp_stb when epp_sel=s_w2o else '0';
	    
  pp:  epp
  	port map (clk_I=> OPB_clk, arst=> arst,
		-- WISHBONE interface
		adr_O=> epp_adr, buf_valid=> epp_buf_valid,
		we_O=> epp_we, cyc_O=> epp_cyc, stb_O=> epp_stb, ack_I=> epp_ack,
		dat_O=> epp_datW, dat_I=> epp_datR, cti_O=> epp_cti,
		-- external (PIN) signals connected to CPLD and Parallel Port (FPGA is a slave device to CPLD!)
		pp_dck=> pp_dck, pp_dwr=> pp_dwr, pp_drd=> pp_drd);
    
     
 epp_rnw<= not epp_we;	
 
 dwidth: epp_dwidth -- convert 8-bit to 16 or 32 bit data
  generic map(dwidth=>c_real_dwidth,  
       use_seqAddr=> c_use_seqAddr)
  port map(clk=> OPB_clk, arst=> arst,
        seqAddr=> epp_cyc, rnw=> epp_rnw, buf_valid=> epp_buf_valid,
        -- 8-bit data interface
        stb8=> w2o_stb, ack8=> w2o_ack, adr8=> epp_adr,
        din8=> epp_datW, dout8=>w2o_dat,
        -- dwidth data interface
        stb=> msel, ack=> mack, be=> mbe, adr=> ma,
        din=> mdrd, dout=> mdwr);

   
  master: opb_master
  generic map(opb_dwidth=> c_opb_dwidth, dwidth=> c_real_dwidth,
    opb_awidth=> c_opb_awidth, awidth=> 32, transfer_size=> 1)
    port map(clk=> OPB_clk, arst=> arst, 
		-- OPB Master interface
	    M_ABus => M_ABus, M_BE=> M_BE, M_RNW=> M_RNW, M_select=> M_select,
	    M_seqAddr=> M_seqAddr, M_busLock=> M_busLock,
	    M_DBus=> M_DBus, M_request=> M_request,
	    OPB_MGrant=> OPB_MGrant, OPB_pendReq=> OPB_pendReq,
	    OPB_DBus=> OPB_DBus, OPB_errAck=> OPB_errAck,
	    OPB_retry=> OPB_retry, OPB_timeout=> OPB_timeout, OPB_xferAck=> OPB_xferAck,
        -- internal simplified bus
        di=> mdwr,  do=> mdrd, be=> mbe, a=> ma, msb_adr=> gnd32,
        rnw=> epp_rnw, sel=> msel, seqAddr=> '0', ack=> mack);
   
  			
---------------------------------------------------------------------------
-- OPB reset (Sys_rst) register
rstp: process(arst, OPB_clk) begin
  if arst='1' then rst_dat(0)<= '0';
  elsif OPB_clk='1' and OPB_clk'event then
	if rst_stb='1' then
		rst_dat(0)<= epp_datW(0);
	end if;
  end if;
 end process;
 
 rst_ack<= rst_stb; -- always ready when selected
 SYS_rst<= rst_dat(0);
 rst_dat(7 downto 1)<= (others=> '0');
 
-----------------------------------------------------------------
-- logic analyser - do not generate logic analyser
  la0: if c_la_mwidth< 4 generate
    la_ack<= '0';
    la_dat<= (others=>'0');
  end generate;
  
  la1: if c_la_mwidth >= 4 generate -- instanciate Logic Analyser
	la: log_anal
      generic map (data_width=> c_la_dwidth, mem_adr_width=> c_la_mwidth, 
	  ce_dwidth=> c_la_ce_dwidth, ced_type=> c_la_ce_dtype,
	    adr_width=> la_awidth, trig_width=> c_la_twidth, two_clocks=> 0, 
        use_run_length_coding=> c_la_run_length_coding)
      port map (arst=> arst, clk=> OPB_clk,
        -- interface for logic analyser 
	    data=> la_data, ce_data=> la_ceDBus,
        trig=> la_trig, ce_trig=> la_ce_trig,
	    -- control WISHBOBE slave interface - interface for setting logic analyser options and transfering analysed data to computer
        wb_clk_I=> OPB_clk, 
	    wb_adr_I=> epp_adr(la_awidth-1 downto 0), wb_dat_I=> epp_datW, wb_dat_O=> la_dat,
	    wb_stb_I=> la_stb, wb_we_I=> epp_we,
	    wb_ack_O=> la_ack); 
    -- standard internal suignals to be analysed by the LA
	g32: if c_la_dwidth<=32 generate
	  la_std_data(7 downto 0)<= OPB_busLock & OPB_seqAddr & OPB_RNW & OPB_errAck 
        & OPB_timeout & OPB_rst & OPB_xferAck & OPB_select;
      g15: if c_la_dwidth=16 and c_la_run_length_coding=1 generate
        la_std_data(14 downto 8)<= OPB_DBus(0 to 6);
      end generate;
      gn15: if c_la_dwidth/=16 or c_la_run_length_coding=0 generate
        la_std_data(15 downto 8)<= OPB_DBus(0 to 7);
      end generate;
      la_std_data(31 downto 16)<= OPB_ABus(c_opb_awidth-16 to c_opb_awidth-1);
	end generate;
	g64: if c_la_dwidth=64 generate
	  la_std_data(7 downto 0)<= OPB_busLock & OPB_seqAddr & OPB_RNW & OPB_errAck 
	      & OPB_timeout & OPB_rst & OPB_xferAck & OPB_select;
      la_std_data(39 downto 8)<= OPB_DBus(0 to 31);
	  la_std_data(63 downto 40)<= OPB_ABus(c_opb_awidth-24 to c_opb_awidth-1);
	end generate;
   
	-- c_la_dtype (data type)
    dtype0: if c_la_dtype=0 generate -- internal logic
	    la_data<= la_std_data(c_la_dwidth-1 downto 0);
	end generate;
	dtype1: if c_la_dtype = 1 generate -- external data
	   la_data<= la_DBus;
	end generate;
	dtype2: if c_la_dtype = 2 generate -- LSB OPB_ABus
	   la_data<= OPB_ABus(c_opb_awidth-c_la_dwidth to c_opb_awidth-1); -- LSBs
	end generate;
	dtype3: if c_la_dtype = 3 generate -- MSB OPB_DBus
	   la_data<= OPB_DBus(0 to c_la_dwidth-1); -- MSBs
	end generate;
    -- c_la_ttype (triger type) -- trigger data type= 0- internal, 1- external, 2- OPB_ABus, 3- OPB_DBus 4- same as la_dbus
	ttype0: if c_la_ttype = 0 generate -- internal (same as internal data)
      la_trig<= la_std_data(c_la_twidth-1 downto 0);
	end generate;
	ttype1: if c_la_ttype = 1 generate -- external
	  la_trig<= la_TBus;
	end generate;
	ttype2: if c_la_ttype = 2 generate -- LSB OPB_ABus
	   la_trig<= OPB_ABus(c_opb_awidth-c_la_twidth to c_opb_awidth-1); -- LSBs
	end generate;
	ttype3: if c_la_ttype = 3 generate -- MSB OPB_DBus
	   la_trig<= OPB_DBus(0 to c_la_twidth-1); -- MSBs
	end generate;
	ttype4: if c_la_ttype = 4 generate -- same as la_DBus
	  la_trig<= la_DBus(0 to c_la_twidth-1);
	end generate;

	-- c_la_tce -- data clock enable= 0- always '1', 1 - external, 2- '1' when OPB_select='1'; 3 - '1' when OPB_xferAck='1', 4 - '1' when OPB_errAck='1' or OPB_xferAck or OPB_retry='1', 5 - OPB_rst
	la_std_ce(0 to 3)<= '1' & la_ceTBus & OPB_select & OPB_xferAck;
	la_std_ce(4)<= OPB_xferAck or OPB_errAck or OPB_retry or OPB_timeout;
    
    la_ce_trig<= la_std_ce(c_la_ce_ttype);
  end generate; -- include logic analyser
	  
-- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
 global_reset: ROC
	port map (O=> arst);
	
 assert c_real_dwidth<=c_opb_dwidth
   report "opb_epp error. c_real_dwidth must be greater than c_opb_dwidth"
   severity failure;
 
end opb_epp_arch;