-- OPB master interface that allows a user logic to be connected to OPB bus as a master
-- this master includes the FIFO

-- Author:  Ernest Jamro (AGH University of Science and Technology - Cracow Poland)
--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement and author name is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////
--//////////////////////////////////////////////////////////////////////



library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

entity opb_master is
  generic(opb_dwidth: integer:= 32; -- opb master data width
    dwidth: integer:= 16; -- internal (real) data width, the LSBs of OPB_DBus='0'
    real_dwidth: integer:= 0; -- opb bus dwidth (=0 then the same as dwidth)
    opb_awidth: integer:= 32; -- opb address bus width
    awidth: integer:= 16; -- internal user address bus width (OPB Master Bus width is greater for read_dwidth<=4 and dwidth/real_dwidth>1
    transfer_size: integer:= 4; -- max transfer size when bus is locked (valid only when seqAddr>0)
    use_be: integer:= 1; -- >0 use Byte Enable Logic, <=0 assume full dwidth transfer all the time
    use_seqAddr: integer:= 1; -- =1 use sequential addressing, =0 -ignore seqAddr signal
    use_addr_counter: integer:= 0; -- <=0 - address is loaded with select signal, >0 use internal address counter, in this mode address (adr) is loaded when load_addr='1' (sel signal is ignored for address logic) and is internally increamented after every data transfer, for reading operation two options are available =1 start reading after load_addr='1', =2 start reading just after reset signal from address=0
    single_master: integer:= 0; -- =1 only a single master is atached to the OPB bus - seting this parameter allows for additional hardware savings, =0 more than one master is connected to the OPB bus
    fifo_wr_awidth: integer:= 0; -- write fifo address width (recommended value = 4) - the size of the fifo= 2^fifo_wr_awidth, =1 - only a single flip-flop; =0 direct connection- fifo is not used, <0 - no connection at all (for write logic)
    fifo_rd_awidth: integer:= 0; -- read fifo address width (same as fifo_wr_awidth but for read logic)
    user_full_transfer: integer:= 1; -- user side full data bus transfer, the parameter is used for  dwidth!= K*real_dwidth or real_dwidth!= K*dwidth
    wr_pipeline_latency: integer:= 0); -- user (external) write logic pipeline latency, (=0 - combinational logic) 1- only a single pipeline register - dwr are delayed by 1 clock in comparison to the address, ack- signal is ack for the address - data (di) are accepted after 1 clk latter (stb and ack signal is disregarded)
  port(clk, arst: std_logic; --clock and global asynchronous reset
    -- OPB Master interface
    M_ABus : out STD_LOGIC_VECTOR(0 TO opb_awidth-1); -- is already gated by M_Select
    M_BE : out STD_LOGIC_VECTOR(0 TO (opb_dwidth+7)/8 -1); 
    M_RNW : out STD_LOGIC;
    M_select : out STD_LOGIC;
    M_seqAddr : out STD_LOGIC;
    M_busLock : out STD_LOGIC;
    M_DBus : out STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
    M_request : out STD_LOGIC;
    OPB_rst: in std_logic:= '0';
    OPB_MGrant: in STD_LOGIC;
    OPB_pendReq: in STD_LOGIC;
    OPB_DBus : in STD_LOGIC_VECTOR(0 TO opb_dwidth-1);
    OPB_errAck : in STD_LOGIC;
    OPB_retry : in STD_LOGIC;
    OPB_timeout: in STD_LOGIC;
    OPB_xferAck : in STD_LOGIC;
    -- internal simplified bus
    load_addr: in std_logic:= '0'; -- load internal address counter - signal is ignored when use_addr_counter<=0
    fifo_full, fifo_empty: out std_logic; -- fifo is full and empty
    fifo_almost_full, fifo_almost_empty: out std_logic; -- only one data can be transferred and fifo will be full or empty
    retry: out std_logic; -- FIFO state indicates that it is better to dwaw retry signal (this signal might be used by an opb_slave module connected to this opb_master module)
    di: in std_logic_vector(0 to dwidth-1):= (others=>'0'); -- data input
    do: out std_logic_vector(0 to dwidth-1); -- data output
    be: in std_logic_vector(0 to (dwidth+7)/8 -1):= (others=>'1'); -- byte enable
    a: in std_logic_vector(0 to awidth-1); -- address bus
    msb_adr: in std_logic_vector(0 to opb_awidth-1):=(others=>'0'); -- the MSB of the OPB address are specified by the msb_adr
    rnw: in std_logic; -- read not write
    sel: in std_logic; -- OPB select signal
    seqAddr: in std_logic; -- seqAddressing - a (address) is incremented - block transfer - similar logic assert OPB_seqAddr
    ack: out std_logic);
end opb_master;


---------------------------------------------------------------------------------------------------------------
--  SRLC (Xilinx build-in shift register)
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;

library unisim;

entity opb_master_srlc is 
  generic(dwidth: integer:= 8; -- data width
    awidth: integer:= 4; -- address width
    use_external_adr: integer:= 0); -- use external address logic
  port (clk, arst, rst: in std_logic; 
    ack_in   : in std_logic; -- input transfer 
    ack_out  : in std_logic; -- output transfer
    adr    : in std_logic_vector(awidth-1 downto 0); -- external address
    din   : in std_logic_vector(dwidth-1 downto 0); 
    dout  : out std_logic_vector(dwidth-1 downto 0)); 
end opb_master_srlc;

architecture syn of opb_master_srlc is 
  component SRL16E
    port ( Q   : out STD_ULOGIC;
      A0  : in STD_ULOGIC;
      A1  : in STD_ULOGIC;
      A2  : in STD_ULOGIC;
      A3  : in STD_ULOGIC;
      CE  : in STD_ULOGIC;
      CLK : in STD_ULOGIC;        
      D   : in STD_ULOGIC
      );
  end component;
  signal adr4: std_logic_vector(3 downto 0);
  signal adr1: std_logic_vector(awidth-1 downto 0); -- internal address
begin 
  adr_ext: if use_external_adr>0 generate
    adr1<= adr;
  end generate;
  
  adr_int: if use_external_adr<=0 generate -- internal address logic
    process(arst, clk) begin
      if arst='1' then adr1<= (others=>'1');
      elsif clk'event and clk='1' then
        if rst='1' then adr1<= (others=>'1');
        elsif ack_in='1' and ack_out='0' then
          adr1<= adr1+1;
        elsif ack_in='0' and ack_out='1' then
          adr1<= adr1-1;
        end if;
      end if;
    end process;
  end generate;
  
  ga3: if awidth<4 generate
    adr4(3 downto 4-awidth)<= (others=>'0');
    adr4(3-awidth downto 0)<= adr1;
  end generate;
  
  ga4: if awidth>=4 generate
    adr4<= adr1(3 downto 0);
  end generate;
  
  gi: for i in 0 to dwidth-1 generate
    gsrl: SRL16E
    port map ( Q=> dout(i),
      A0=> adr4(0), A1=> adr4(1),
      A2=> adr4(2), A3=> adr4(3),
      CE=> ack_in, CLK=> clk, D=> din(i)); 
  end generate;
end syn;


---------------------------------------------------------------------------------------------------------------
-- multiplexer
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;

entity opb_master_mux is 
  generic(din_width: integer:= 20; -- input width width
    dout_width: integer:= 8; -- data out width
    sel_width: integer:=2; -- select signal width
    K: integer:= 3); -- address width
  port (sel: in std_logic_vector(0 to sel_width-1);
    din: in std_logic_vector(0 to din_width-1);
    dout: out std_logic_vector(0 to dout_width-1));
end opb_master_mux;

architecture synth of opb_master_mux is
  function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MAX;
  type mux_din_type is array (0 to K-1) of std_logic_vector(0 to dout_width-1);
  signal mux_din: mux_din_type;
  constant din1_width: integer:= K*dout_width;
  type dout_and_type is array (0 to dout_width-1) of std_logic_vector(0 to K-1);
  signal dout_and: dout_and_type;
  type mux4_type is array (0 to K/4) of std_logic_vector(0 to dout_width-1); -- 4:1 multiplexer output
  signal mux4_dout: mux4_type;
  type mux5_type is array (0 to dout_width-1) of std_logic_vector(0 to K/4); -- additional multiplexer for K>4
  signal mux5_and: mux5_type;
begin
  gmux_in: for i in 0 to K-1 generate
    g0: if (K-i)*dout_width<= din_width generate
      mux_din(i)<= din(din_width - (K-i)*dout_width to din_width-(K-1-i)*dout_width-1);
    end generate;
    g1: if (K-i)*dout_width > din_width generate -- only for i=0
      mux_din(0)(0 to K*dout_width-din_width-1)<= (others=>'0');
      mux_din(0)(K*dout_width-din_width to dout_width-1)<= 
      din(0 to din_width - (K-1)*dout_width -1);
    end generate;
  end generate;
  
  gmux4: if K>=4 generate
    gmux4: for i in 0 to K/4 -1 generate
      mux4_dout(i)<= mux_din(4*i) when sel(sel_width-2 to sel_width-1)="00" else
      mux_din(4*i+1) when sel(sel_width-2 to sel_width-1)="01" else
      mux_din(4*i+2) when sel(sel_width-2 to sel_width-1)="10" else
      mux_din(4*i+3);
    end generate;
  end generate;
  
  --  gmux3: if K rem 4 = 3 generate
  --     mux4_dout(K/4)<= (others=>'0') when sel(sel_width-2 to sel_width-1)="00" else
  --                  mux_din(K-3) when sel(sel_width-2 to sel_width-1)="01" else
  --                  mux_din(K-2) when sel(sel_width-2 to sel_width-1)="10" else
  --                  mux_din(K-1);
  --  end generate;
  
  gmux2: if K rem 4 = 2 generate
    mux4_dout(K/4)<= mux_din(K-2) when sel(sel_width-1)='0' else
    mux_din(K-1);
  end generate;
  
  gmux1: if K rem 4 = 1 generate
    mux4_dout(K/4)<= mux_din(K-1);
  end generate;
  
  gK4: if K<=4 generate
    dout<= mux4_dout(0);
  end generate;
  
  -- additional multiplexer for K>4
  gK5: if K>4 generate
    gi: for i in 0 to dout_width-1 generate
      gj: for j in 0 to K/4 generate
        mux5_and(i)(j)<= mux4_dout(j)(i) when sel(0 to sel_width-3)=j else '0';
      end generate;
      dout(i)<= OR_REDUCE(mux5_and(i));
    end generate;
  end generate;
  
  
  --g0: if din1_width>din_width generate
  --  din1(din1_width-din_width to din1_width-1)<= din; -- align to the LSBs
  --  gdin1: if din1_width>din_width generate -- extend the input width zeros
  --    din1(0 to din1_width-din_width-1)<= (others=>'0');
  --  end generate;
  --  
  --  gand: for i in 0 to K-1 generate -- multiplexer and gate
  --    din1_and(i*dout_width to (i+1)*dout_width-1)<= 
  --      din1(i*dout_width to (i+1)*dout_width-1) when sel= (2**sel_width)-K+i -- align to the LSBs
  --    4  else (others=>'0');
  --  end generate;  
  --  
  --  garr: for i in 0 to dout_width-1 generate -- rearrange the din1_and vector for proper OR function
  --    gj: for j in 0 to K-1 generate
  --      dout_and(i)(j)<= din1_and(j*dout_width + i);
  --    end generate;
  --  end generate;
  --
  --  gor: for i in 0 to dout_width-1 generate -- or gate
  --    dout(i)<= OR_REDUCE(dout_and(i));
  --  end generate;
end synth;


-----------------------------------------------------------------------------------------------------
-- this fifo can work only for din_width<= dout_width

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;

entity opb_master_fifo is
  generic(fifo_awidth: integer:= 4; -- internal FIFO memory address width - 4 is preferable for distributed rams, 1- only a singlr Flip-flop, 0- direct connection, <0- do not implement fifo logic at all
    awidth: integer:= 1; -- address width
    dout_width: integer; -- data output width 
    din_width: integer; -- data input width din_width <= dout_width
    use_be: integer;
    use_seqAddr: integer;
    Kin: integer; -- (dout_width(+din_width-1)) / din_width; see user_full_transfer
    transfer_size: integer:= 1); -- the transfer size (same as opb_master)
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
    rst: in std_logic; -- synchronous reset for the whole device
    transfer_full, transfer_empty: out std_logic; -- transfer_full = 1 - fifo may not accept transfer_size data, transfer_empty - fifo contains less than transfer_size data
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    -- data input signals (slave)
    sel_in: in std_logic;
    ack_in: out std_logic;
    adr_in: in std_logic_vector(0 to awidth-1); -- address - used only for din_width<dout_width
    din: in std_logic_vector(0 to din_width-1);
    bin: in std_logic_vector(0 to (din_width+7)/8-1); -- input byte enable
    -- data output signals (master)
    sel_out: out std_logic; -- at least one data are valid
    sel_out_full: out std_logic; -- full word output transfer is available (no more concatenations is possible)
    ack_out: in std_logic; -- full output transfer is always assumed
    dout: out std_logic_vector(0 to dout_width-1);
    bout: out std_logic_vector(0 to (dout_width+7)/8-1)); --(bout - output byte enable
end opb_master_fifo;

architecture arch of opb_master_fifo is
  component opb_master_srlc
    generic(dwidth: integer:= 8; -- data width
      awidth: integer:= 4; -- address width
      use_external_adr: integer:= 0); -- use external address logic
    port (clk, arst, rst: in std_logic; 
      ack_in   : in std_logic; -- input transfer 
      ack_out  : in std_logic; -- output transfer
      adr    : in std_logic_vector(awidth-1 downto 0); -- external address
      din   : in std_logic_vector(dwidth-1 downto 0); 
      dout  : out std_logic_vector(dwidth-1 downto 0)); 
  end component;
  
  -- find maximum value of two integers
  function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MAX;
  function MIN(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0<=arg1 then ret:= arg0;
    else ret:= arg1;
    end if;
    return ret;
  end MIN;
  
  -- find fifo memory data width
  constant dout_width1: integer:= MIN(Kin*din_width, dout_width); -- in some cases not all dout data are used
  
  -- find Byte Enable length (the byte enable length can be different from 8 - however in this case use_be<0
  function FindDBeWidth(ignore: integer) return integer is begin
    if use_be<=0 then
      return din_width;
    else
      return 8;
    end if;
  end FindDBeWidth;
  constant dbe_width: integer:= FindDBeWidth(0); -- number of bits for a single BYTE ENABLE logic
  
  constant bin_width: integer:= (din_width+dbe_width-1)/dbe_width; -- real input byte enable width
  constant bout_width: integer:= (dout_width1+dbe_width-1)/dbe_width; -- output byte enable
  constant fifo_awidth1: integer:= MAX(1, fifo_awidth); -- address width for signal declaration
  constant sh_dwidth: integer:= din_width + MAX(0, MIN(1, use_be))*bin_width; -- shift register data width (include byte enable if required)
  signal sh_din, sh_dout: std_logic_vector(0 to Kin*sh_dwidth-1); -- shift reg. data out 
  signal empty: std_logic; 
  --signal count_in, count_out: std_logic_vector(fifo_awidth1-1 downto 0); -- counter where place data in into the memory or from where read data out
  signal count_dif: std_logic_vector(fifo_awidth1-1 downto 0); -- counter how many data are inside the buffer = count_in-count_out-1
  signal ackI, ackO: std_logic; -- input/output  ack
  signal ackI_adr: std_logic; -- next address input
  signal valid_din: std_logic; -- valid data in the input fifo - but fifo has not yet increase the count_in (not full words transfer finished)
  signal bin1, beq, ram_we, ram_be_we: std_logic_vector(0 to bout_width-1); -- be1- byte enable same width as bout, beq- byte enable flip-flops used only for fifo_awidth=1
  signal sh_we, sh_weq: std_logic_vector(0 to Kin-1); -- byte enable for each input transfer (only a single for all input data)
begin
  -- bin1 logic - byte enable input extended to bout_width - use address logic - align to the LSB
  fbe1: for i in 0 to bout_width-1 generate
    bin1(i)<= '0' when adr_in /= ((2**awidth) -1 - (bout_width-1-i)/bin_width) and Kin>1 else -- proper address
    '1' when use_be<=0 else 
    bin(bin_width-1 - ((bout_width-1-i) rem bin_width));
  end generate;
  
  gsh_we: for i in 0 to Kin-1 generate -- shift register write
    sh_we(i)<= ackI when Kin=1 or (adr_in = ((2**awidth) - Kin + i)) else -- check for proper address
    ackI_adr when sh_weq(i)='0' and i<Kin-1 and use_be>0 else '0'; -- load shift register to set BE to zero
  end generate;  
  ackI_adr<= sh_we(Kin-1); -- the last transfer
  
  -- use FIFO memory
  gadr2: if fifo_awidth> 1 and use_seqAddr>0 generate
    -- input counter
    process(clk, arst) begin
      if arst='1' then valid_din<= '0';
      elsif clk'event and clk='1' then
        
        if Kin>1 and use_be>0 then  -- valid_din logic shows that there are valid (not full) input data
          if rst='1' or ackI_adr='1' then  -- ackI_adr - full input transfer so reset
            valid_din<= '0'; 
          elsif ackI='1' then -- partial input transfer
            valid_din<= '1'; -- valid data in the FIFO but fifo count_in is still all (not whole input data write)
          elsif ack_out='1' and empty='1' then -- the last (not full) output transfer
            valid_din<= '0';
          end if;
        end if;
      end if;
    end process;
    
    ackI<= sel_in and not full; -- no input handshaking
    ack_in<= ackI;
    sel_out<= not empty or valid_din when use_be>0 and Kin>1 else not empty;
    sel_out_full<= not empty;
    ackO<= ack_out when Kin=1 or use_be<=0 else ack_out and not empty; -- in the case when not full out transfer occures the fifo counters should not be changed
    
    -- fifo counter - how many data are in the fifo (count_dif)
    -- count_dif<= count_in - count_out-1; 
    c_dif: process(clk, arst) begin
      if arst='1' then count_dif<= (others=>'1');
      elsif clk'event and clk='1' then 
        if rst='1' then count_dif<= (others=>'1');
        elsif ackI_adr='1' and ackO='0' then
          count_dif<= count_dif + 1;
        elsif ackI_adr='0' and ackO='1' then
          count_dif<= count_dif - 1;
        end if;
      end if;
    end process;
    -- empty logic
    process(clk, arst) begin
      if arst='1' then empty<= '1'; almost_empty<= '1'; almost_full<= '0'; transfer_empty<= '0'; 
        transfer_full<= '0'; 
      elsif clk'event and clk='1' then
        
        if rst='1' then empty<= '1';
        elsif ackI_adr='1' then
          empty<= '0';
        elsif count_dif=0 and ackO='1' then
          empty<= '1';
        end if;
        
        -- almost_empty logic
        if empty='1' or count_dif=0 or (count_dif=1 and ackO='1') then -- assume that ackI='0'
          almost_empty<= '1';
        else almost_empty<= '0';
        end if;
        -- almost full logic
        if rst='1' or empty='1' or          
          (Kin=1 and count_dif < (2**fifo_awidth)-2) or 
          ((Kin>1 and ackO='1')) then
          almost_full<= '0';
        elsif Kin=1 then
          almost_full<= '1';
        elsif count_dif = 2**fifo_awidth-2 and OR_REDUCE(bin1((Kin-1)*bin_width to bout_width-1))='1' then
          almost_full<= '1';
        end if;
        -- transfer_full
        if count_dif <= (2**fifo_awidth)- 1 - (transfer_size+Kin-1)/Kin or empty='1' then
          transfer_full<= '0';
        else transfer_full<='1';
        end if;
        -- transfer empty
        if count_dif >= transfer_size-1 and empty='0' then
          transfer_empty<= '0';
        else transfer_empty<='1';
        end if;
      end if;
    end process;
    
    -- full logic - full should be '1' whenever the count_dif = 2**adr_width
    process(clk, arst) begin
      if arst='1' then full<= '0';
      elsif clk'event and clk='1' then
        if rst='1' then full<= '0';
        elsif (full='1' and ackO='0') or -- already full and cannot send data out (do not accept new data when full)
          (count_dif=(2**fifo_awidth)-2 and ackO='0' and ackI_adr='1' ) then -- almost full and new data will be accepted in and no data send out
          full<= '1';
        else full<= '0';
        end if;
      end if; 
    end process;  
    
    -- internal memory SRL (Xilinx shift register)
    shi: for i in 0 to Kin-1 generate
      shift: opb_master_srlc
      generic map(dwidth=> sh_dwidth,
        awidth=> fifo_awidth,
        use_external_adr=> i+2-Kin) -- use external address only for the last module
      port map(clk=> clk, arst=>arst, rst=>rst,
        ack_in=> sh_we(i), ack_out=> ackO,
        adr=> count_dif, din=> sh_din(i*sh_dwidth to (i+1)*sh_dwidth-1),
        dout=> sh_dout(i*sh_dwidth to (i+1)*sh_dwidth-1));
      
      sh_din(i*sh_dwidth to i*sh_dwidth + din_width-1)<= din;
      
      gbe1: if use_be>0 generate -- use Byte Enable logic so Kin must be an integer=> dout_width = Kin*din_width
        sh_din(i*sh_dwidth+din_width to (i+1)*sh_dwidth-1) <= -- byte enable
        bin1(i*bin_width to (i+1)*bin_width-1); 
        dout(i*din_width to (i+1)*din_width-1)<= sh_dout(i*sh_dwidth to i*sh_dwidth+din_width-1);
        bout(i*bin_width to (i+1)*bin_width-1)<= sh_dout(i*sh_dwidth+din_width to (i+1)*sh_dwidth-1) 
        when empty='0' or sh_weq(i)='1' else (others=>'0'); -- valid data in shift register
      end generate;
    end generate;
    -- write enable logic (remember each byte enable for not full transfer)
    process(arst, clk) begin
      if arst='1' then sh_weq<= (others=>'0');
      elsif clk'event and clk='1' then
        if use_be>0 and Kin>1 then
          for i in 0 to Kin-2 loop
            if rst='1' or ackI_adr='1' then 
              sh_weq(i)<= '0';
            elsif sh_we(i)='1' then
              sh_weq(i)<= '1';
            end if;
          end loop;
        end if;
      end if;
    end process;  
    
    sh_weq(Kin-1)<= '0'; -- the last 
    gsh_dout: if use_be<=0 generate -- do not use BE so it is possible that dout_width != Kin*din_width
      ga: if Kin*din_width< dout_width generate
        dout(dout_width - Kin*din_width to dout_width-1)<= sh_dout; -- align to the LSBs
        dout(0 to dout_width-Kin*din_width-1)<= (others=>'0'); -- fill MSBs with zeros
      end generate;
      gb: if Kin*din_width>=dout_width generate
        dout<= sh_dout(Kin*din_width-dout_width to Kin*din_width-1);
      end generate;
    end generate;             
  end generate; -- adr_width>= 1
  
  -----------------------------------------------------------------------------------------------------------
  
  -- only a single flip-flop register
  gadr1: if fifo_awidth= 1 or (fifo_awidth>1 and use_seqAddr<=0) generate
    ackI<= not full and sel_in;
    ack_in<= ackI;
    sel_out<= '1' when full='1' or (bout_width>1 and valid_din='1') else '0';
    sel_out_full<= full; 
    
    process(clk, arst) begin
      if arst='1' then dout<= (others=> '0'); valid_din<= '0'; full<= '0'; beq<= (others=>'0');
      elsif clk'event and clk='1' then
        for i in 0 to bout_width-1 loop
          if ackI='1' and bin1(i)='1' then
            dout(MAX(0, dout_width1-(bout_width-i)*dbe_width) 
            to dout_width1-(bout_width-i-1)*dbe_width -1) <= 
            din( (Kin*din_width - MIN(dout_width1, (bout_width-i)*dbe_width)) rem din_width 
            to (Kin*din_width - (bout_width-i-1)*dbe_width-1) rem din_width);
          end if;
          if use_be>0 and ackI='1' then -- use byte enable 
            if valid_din='0' then -- first transfer
              beq(i)<= bin1(i);
            elsif bin1(i)='1' then
              beq(i)<= '1';
            end if;
          end if;
        end loop;
        if rst='1' or ack_out='1' then
          valid_din<= '0'; full<= '0';
        elsif ackI='1' and OR_REDUCE(bin1((Kin-1)*bin_width to bout_width-1))='1' then -- the last cycle 
          full<= '1'; valid_din<='0';
        elsif bout_width>1 and ackI='1' then 
          valid_din<= '1';
        end if;
      end if;
    end process;
    gb1: if use_be>0 generate
      bout(0 to bout_width-1)<= beq;
    end generate;
    almost_full<= '1' when Kin=1 else 
    OR_REDUCE(beq((Kin-1)*bin_width to bout_width-1)) when use_be>0 else
    '1' when adr_in>= (2**awidth)-2 else '0';  
    gtf1: if Kin<=transfer_size generate
      transfer_full<= '1';
    end generate;
    gtf0: if Kin > transfer_size generate
      transfer_full<= OR_REDUCE(beq((Kin-transfer_size)*bin_width to bout_width-1)) when use_be>0 else
      '1' when adr_in>= (2**awidth)-transfer_size else '0';
    end generate;
    almost_empty<= '1'; 
    transfer_empty<= '1';
  end generate; 
  
  ---------------------------------------------------------------
  
  gadr0: if fifo_awidth=0 generate -- use direct connection
    -- directly copy input data to output
    gi: for i in 0 to Kin-1 generate
      gd0: if (i+1)*din_width<= dout_width generate
        dout(dout_width - (i+1)* din_width to dout_width - i*din_width -1)<= din;
        gbe: if use_be>0 generate
          bout(bout_width - (i+1)*bin_width to bout_width - i*bin_width-1)<= bin;
        end generate;
      end generate;
      gd1: if (i+1)*din_width> dout_width generate
        dout(0 to (i+1)*din_width - dout_width-1)<= 
        din(din_width - ((i+1)*din_width - dout_width) to din_width-1);
      end generate;
    end generate;
    
    sel_out<= sel_in;
    sel_out_full<= sel_in;
    ack_in<= ack_out;
    almost_full<= '1';
    almost_empty<= '1';
    transfer_full<= '1';
    transfer_empty<= '1';
    empty<= '0';
    full<= '0' ; -- not sel_in; -- this logic is required for reading: msel<= not fifo_full -- fifo is full when not ack
    gbe1: if use_be>0 generate
      bout(0 to bout_width-1)<= bin1;
    end generate;
  end generate;
  
  gadr_l0: if fifo_awidth<0 generate -- not logic at all
    dout<= (others=> '0');
    sel_out<= '0';
    sel_out_full<= '0';
    ack_in<='0';
    almost_full<= '1';
    almost_empty<= '1';
    full<= '1';
    bout<= (others=> '0');
  end generate;
  
  gd1: if dout_width>dout_width1 and fifo_awidth>=0 generate
    dout(dout_width1 to dout_width-1)<= (others=>'0');
  end generate;
  gbe0: if use_be<=0 and fifo_awidth>=0 generate
    bout<= (others=>'1');
  end generate;
  gbe1: if (dout_width+7)/8 > bout_width and use_be>0 and fifo_awidth>=0 generate
    bout(bout_width to (dout_width+7)/8 -1)<= (others=>'0');
  end generate;
  
  assert fifo_awidth<=5
  report "opb_master (slave)FIFO. This architecture does not support fifo_awidth>4"
  severity error; 
end arch;                


-----------------------------------------------------------------------------------------------------
-- 2 directions FIFO (seperate fifo for reads and writes) - for mdwidth = dwidth only a single FIFO is used for read and writes

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;


entity opb_master_fifo2 is
  generic(awidth: integer:= 4; -- user address bus width - LSBs of the address bus are taken into account
    mawidth: integer:= 4; -- master side address width - the same as awidth for dwidth>4
    dwidth: integer:= 8; -- user data width
    mdwidth: integer:= 16; -- opb Bus data width
    use_be: integer:= 1; -- use byte enable
    use_seqAddr: integer:= 1; -- use sequantiaal addressing
    user_full_transfer: integer; 
    fifo_wr_awidth: integer:= 4; -- write fifo address width (rnw=0)
    fifo_rd_awidth: integer:= 4; -- read fifo address width
    wr_pipeline_latency: integer:= 0; 
    transfer_size: integer:= 4); -- the transfer size (same as opb_master)
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
    rst: in std_logic; -- synchronous reset for the whole device
    rnw: in std_logic; -- read not write operation
    transfer_full, transfer_empty: out std_logic; -- transfer_full = 1 - fifo may not accept transfer_size data, transfer_empty - fifo contains less than transfer_size data
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    empty: out std_logic; -- empty (no data in the buffer)
    finish: in std_logic; -- finish - generate msel for write even when not full output word is valid
    -- user interface
    adr: in std_logic_vector(0 to awidth-1);
    dwr: in std_logic_vector(0 to dwidth-1);
    drd: out std_logic_vector(0 to dwidth-1);
    sel: in std_logic;
    seqAddr: in std_logic;
    ack: out std_logic;
    be: in std_logic_vector(0 to (dwidth+7)/8-1);
    -- data output signals (master)
    mseqAddr_in: in std_logic; -- seqAddressing - generated in opb_master_aux
    madr_out: out std_logic_vector(0 to mawidth-1);
    madr_in: in std_logic_vector(0 to mawidth-1);
    mdwr: out std_logic_vector(0 to mdwidth-1);
    mdrd: in std_logic_vector(0 to mdwidth-1);
    mbe: out std_logic_vector(0 to (mdwidth+7)/8-1);
    msel: out std_logic;
    mack: in std_logic);
end opb_master_fifo2;

architecture arch of opb_master_fifo2 is
  component opb_master_mux
    generic(din_width: integer:= 20; -- input width width
      dout_width: integer:= 8; -- data out width
      sel_width: integer:=2; -- select signal width
      K: integer:= 3); -- address width
    port (sel: in std_logic_vector(0 to sel_width-1);
      din: in std_logic_vector(0 to din_width-1);
      dout: out std_logic_vector(0 to dout_width-1));
  end component;
  component opb_master_fifo
    generic(fifo_awidth: integer:= 4; -- internal FIFO memory address width - 4 is preferable for distributed rams, 1- only a singlr Flip-flop, 0- direct connection, <0- do not implement fifo logic at all
      awidth: integer:= 1; -- address width
      dout_width: integer; -- data output width 
      din_width: integer; -- data input width din_width <= dout_width
      use_seqAddr: integer;
      use_be: integer;
      Kin: integer; -- (dout_width(+din_width-1)) / din_width; see user_full_transfer
      transfer_size: integer:= 1); -- the transfer size (same as opb_master)
    port (clk: in std_logic; -- clock
      arst: in std_logic; -- global asynchronous reset
      rst: in std_logic; -- synchronous reset for the whole device
      transfer_full, transfer_empty: out std_logic; -- transfer_full = 1 - fifo may not accept transfer_size data, transfer_empty - fifo contains less than transfer_size data
      almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
      full: buffer std_logic;
      -- data input signals (slave)
      sel_in: in std_logic;
      ack_in: out std_logic;
      adr_in: in std_logic_vector(0 to awidth-1); -- address - used only for din_width<dout_width
      din: in std_logic_vector(0 to din_width-1);
      bin: in std_logic_vector(0 to (din_width+7)/8-1); -- input byte enable
      -- data output signals (master)
      sel_out: out std_logic;
      sel_out_full: out std_logic; -- full word output transfer is available (no more concatenations is possible)
      ack_out: in std_logic; -- full output transfer is always assumed
      dout: out std_logic_vector(0 to dout_width-1);
      bout: out std_logic_vector(0 to (dout_width+7)/8-1)); 
  end component;
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
    for i in 0 to 32 loop
      if 2**i >= arg then ret:= i; exit; end if;
    end loop;
    return ret;
  end log2;
  function MAX(arg0: integer; arg1: integer) return integer is begin
    if arg0>=arg1 then
      return arg0;
    else 
      return arg1;
    end if;
  end MAX;
  function MIN(arg0: integer; arg1: integer) return integer is begin
    if arg0<=arg1 then
      return arg0;
    else 
      return arg1;
    end if;
  end MIN;
  function FindFifoAwidth(awidth0:integer; awidth1: integer) return integer is begin
    if awidth0>0 then
      if use_seqAddr>0 then
        return awidth0;
      else return 1; -- limit fifo Awidth to 1 when seqAddr<=0
      end if;
    elsif awidth0=0 then
      if use_seqAddr>0 and dwidth/=mdwidth then
        return 4; -- use FIFO for quick data width conversion
      elsif awidth1>0 then  -- cannot use direct connection when other connection is not direct
        return 1;
      else return 0;
      end if;
    else return awidth0; -- <0
    end if;
  end FindFifoAwidth;
  
  constant fifo_wr_awidth1: integer:= FindFifoAwidth(fifo_wr_awidth, fifo_rd_awidth);
  constant fifo_rd_awidth1: integer:= FindFifoAwidth(fifo_rd_awidth, fifo_wr_awidth);
  -- find if the Byte enable logic should be used or Bete enable should be ignored
  function FindUseBe(din_width: integer; dout_width: integer) return integer is begin
    if din_width<8 or dout_width<=8 or use_be<=0 or 
      ( (din_width /= dout_width) and  (din_width rem 8)/=0 ) then -- no Byte Enable logic possible when din_width rem 8 !=0
      return 0;
    else
      return 1;
    end if;
  end FindUseBe;
  
  constant use_be_wr: integer:= FindUseBe(dwidth, mdwidth);
  constant K: integer:= MAX((dwidth+user_full_transfer*(mdwidth-1))/mdwidth, -- rounding down
  (mdwidth + (1-user_full_transfer)*(dwidth-1))/dwidth); -- number of transfers (rounding up)
  constant mbe_width: integer:= (mdwidth+7)/8; -- master Byte Enable width
  constant lsb_awidth: integer:= awidth-log2((dwidth+7)/8); -- LSBs which are taken into account by fifo (user side)
  constant lsb_mawidth: integer:= mawidth-log2((mdwidth+7)/8); -- LSBs which are taken into account by fifo (master side)
  constant be_awidth: integer:= MAX(1, MAX(lsb_awidth-lsb_mawidth, lsb_mawidth-lsb_awidth)); -- address width used by fifo
  constant bin_rd_ones: std_logic_vector(0 to (mdwidth+7)/8-1):= (others=>'1');
  constant one: std_logic_vector(0 to 0):= "0";
  signal fwr_sel_in, fwr_sel_out, fwr_sel_out_full, fwr_ack_in, fwr_ack_out: std_logic;
  signal fwr_bout, frd_bout, mbe1: std_logic_vector(0 to (mdwidth+7)/8-1);
  signal fwr_boutM: std_logic_vector(0 to (dwidth+7)/8-1); -- internal mbe used for dwidth>mdwidth
  signal frd_sel_in, frd_sel_out_full, frd_ack_out: std_logic;
  signal fdin, fdout: std_logic_vector(0 to dwidth-1); -- used only for K=1
  signal fdrd: std_logic_vector(0 to mdwidth-1); -- data used only for mdwidth>dwidth
  signal fwr_almost_full, fwr_almost_empty, fwr_transfer_full, fwr_full: std_logic;
  signal frd_almost_full, frd_almost_empty, frd_transfer_full, frd_full: std_logic;
  signal reading: std_logic; -- reading operation is performed
  signal msel1, mseqAddr_QN: std_logic; -- logic that is used to sepereate 2 blocks that are not seqAddressed
begin
  
  assert K=1 or K=2 or K=4 or K=8 or K=16 or K=32 
  report "opb_master, Data width conversion wrong parameters. For a single master (user) side transfer, the number of transfers on master (user) size must be a power of 2"
  severity failure;
   
  msel<= msel1; 
   -- logic required to seperate two seq Addressing blocks - in some cases two different blocks aremerged together which cuases errors
  gsa1: if (fifo_rd_awidth>0 or (fifo_rd_awidth=0 and fifo_wr_awidth>0)) and use_seqAddr>0 generate -- when using read register
    process (arst, clk) begin
      if arst='1' then mseqAddr_QN<= '0';
      elsif clk'event and clk='1' then
        if (mseqAddr_in='0' and mack='1') or msel1='0' then -- end of seqAddressing
          mseqAddr_QN<= '0';
        elsif seqAddr='0' and sel='0' then
          mseqAddr_QN<= '1'; --end previous seqAddressong block
        end if;
      end if;
    end process;
  end generate;
  
  gsa0: if not ((fifo_rd_awidth>0 or (fifo_rd_awidth=0 and fifo_wr_awidth>0)) and use_seqAddr>0) generate -- when using read register
    mseqAddr_QN<= '0';
  end generate;
    
  reading<= '0' when fifo_rd_awidth<0 else rnw; -- read operation
  
  Kmaster: if dwidth>mdwidth generate -- several master transfers for one user transfer
    fifo_wr: opb_master_fifo
    generic map(fifo_awidth=> fifo_wr_awidth1, awidth=> 1, Kin=> 1,
      dout_width=> dwidth, din_width=> dwidth, use_be=> use_be_wr,
      use_seqAddr=> use_seqAddr, transfer_size=> transfer_size+wr_pipeline_latency)
    port map(clk=> clk, arst=> arst, rst=> rst,
      transfer_full=> fwr_transfer_full, transfer_empty=> transfer_empty,
      almost_full=> fwr_almost_full, almost_empty=>fwr_almost_empty, full=> fwr_full,
      -- data input signals (slave)
      sel_in=> fwr_sel_in, ack_in=> fwr_ack_in, 
      adr_in=> one,     
      din=> dwr, bin=> be,
      -- data output signals (master)
      sel_out=> open, sel_out_full=> fwr_sel_out_full, ack_out=> fwr_ack_out,
      dout=> fdout, bout=> fwr_boutM);
    fifo_rd: opb_master_fifo
    generic map(fifo_awidth=> fifo_rd_awidth1, awidth=> be_awidth, use_be=> 0,
      dout_width=> dwidth, din_width=> mdwidth, Kin=> K,
      use_seqAddr=> 1, transfer_size=> (transfer_size+K-1)/K)
    port map(clk=> clk, arst=> arst, rst=> rst,
      transfer_full=> frd_transfer_full, transfer_empty=> open,
      almost_full=> frd_almost_full, almost_empty=>frd_almost_empty, full=> frd_full,
      -- data input signals (slave)
      sel_in=> frd_sel_in, ack_in=> open, adr_in=> madr_in(lsb_awidth to lsb_mawidth-1),
      din=> mdrd, bin=> bin_rd_ones,
      -- data output signals (master)
      sel_out=> open, sel_out_full=> frd_sel_out_full, ack_out=> frd_ack_out,
      dout=> drd, bout=> open);
    
    empty<= '1' when fifo_wr_awidth<=0 and fifo_rd_awidth<=0 else -- direct connection - empty logic is not valid
    not frd_sel_out_full when reading='1' else not fwr_sel_out_full; 
    full<= frd_full when reading='1' else fwr_full;
    almost_full<= frd_almost_full when reading='1' else fwr_almost_full;
    almost_empty<= not frd_sel_out_full when reading='1' else not fwr_sel_out_full; -- almost empty logic does not work correctly for this configuration !!!
    transfer_full<= frd_transfer_full when wr_pipeline_latency<=0 or rnw='1' else fwr_transfer_full; -- transfer FULL is required only reading or for writes when wr_pipeline_latency>0
    ack<= fwr_ack_in when reading='0' else sel and frd_sel_out_full;
    msel1<= not frd_full and not mseqAddr_QN when reading='1' and (fifo_wr_awidth>0 or fifo_rd_awidth>0) else
      sel when reading='1' else fwr_sel_out_full;
    fwr_sel_in<= sel and not reading;
    -- fwr_ack_out logic
    process(mack, reading, madr_in(lsb_awidth to lsb_mawidth-1), fwr_boutM) begin
      if mack='0' or reading='1' then fwr_ack_out<= '0';
      elsif madr_in(lsb_awidth to lsb_mawidth-1) = (2**be_awidth)-1 then fwr_ack_out<= '1';-- the last write
      elsif use_be_wr<=0 then fwr_ack_out<= '0'; -- only full transfers are considered for use_be<=0
      elsif madr_in(lsb_awidth to lsb_mawidth-1)= 2**be_awidth-2 and OR_REDUCE(fwr_boutM((K-1)*mbe_width to K*mbe_width-1))='0' then -- the last transfer - the next LSB BE bits are '0' so this transfer is the last transfer
        fwr_ack_out<= '1'; -- the next transfer(s) is not valid
      elsif K>2 and madr_in(lsb_awidth to lsb_mawidth-1)=2**be_awidth-3 and OR_REDUCE(fwr_boutM((K-2)*mbe_width to K*mbe_width-1))='0'then
        fwr_ack_out<= '1'; -- the next transfer(s) is not valid
      elsif K>3 and madr_in(lsb_awidth to lsb_mawidth-1)=2**be_awidth-4 and OR_REDUCE(fwr_boutM((K-3)*mbe_width to K*mbe_width-1))='0' then
        fwr_ack_out<= '1'; -- the next transfer(s) is not valid
      else fwr_ack_out<= '0';
      end if;
    end process;
    frd_sel_in<= mack and reading;
    frd_ack_out<= sel and frd_sel_out_full;
    -- write data multiplexer
    mux: opb_master_mux
    generic map(din_width=> dwidth, dout_width=> mdwidth,
      sel_width=> be_awidth, K=> K)
    port map(sel=> madr_in(lsb_awidth to lsb_mawidth-1), 
      din=> fdout, dout=> mdwr);
    mux_be: if use_be_wr>0 generate
      muxbe: opb_master_mux
      generic map(din_width=> (dwidth+7)/8, dout_width=> (mdwidth+7)/8,
        sel_width=> be_awidth, K=> K)
      port map(sel=> madr_in(lsb_awidth to lsb_mawidth-1), 
        din=> fwr_boutM, dout=> fwr_bout);
    end generate;
    mbe1<= (others=>'1') when use_be_wr<=0 or reading='1' else fwr_bout;         
    --    process(madr_in(lsb_awidth to lsb_mawidth-1), fdout, fwr_boutM, reading) begin 
    --      for i in 0 to K-1 loop
    --        if madr_in(lsb_awidth to lsb_mawidth-1)=(2**be_awidth)-K+i then
    --          mdwr(0 to MIN(mdwidth, dwidth - i*mdwidth)-1)<= 
    --          fdout(dwidth - (K - i)*mdwidth to MIN(dwidth, dwidth - (K-i-1)*mdwidth)-1);
    --          if (i+1)*mdwidth > dwidth then
    --            mdwr(dwidth - i*mdwidth to mdwidth-1)<= (others=>'0');
    --          end if;
    --          if reading='1' or use_be_wr<=0 then 
    --            mbe1<= (others=>'1');
    --          else
    --            mbe1<= fwr_boutM(i*((mdwidth+7)/8) to (i+1)*((mdwidth+7)/8)-1);
    --          end if;
    --        end if;
    --      end loop;
    --    end process;
  end generate;
  
  
  Kuser: if dwidth<mdwidth generate -- several user transfers for one master transfer
    fifo_wr: opb_master_fifo
    generic map(fifo_awidth=> fifo_wr_awidth1, awidth=> be_awidth, Kin=> K,
      dout_width=> mdwidth, din_width=> dwidth, use_be=> use_be_wr,
      use_seqAddr=> use_seqAddr, transfer_size=> transfer_size+wr_pipeline_latency)
    port map(clk=> clk, arst=> arst, rst=> rst,
      transfer_full=> fwr_transfer_full, transfer_empty=> transfer_empty,
      almost_full=> fwr_almost_full, almost_empty=>fwr_almost_empty, full=> fwr_full,
      -- data input signals (slave)
      sel_in=> fwr_sel_in, ack_in=> fwr_ack_in, 
      adr_in=> adr(lsb_mawidth to lsb_awidth-1),
      din=> dwr, bin=> be,
      -- data output signals (master)
      sel_out=> fwr_sel_out, sel_out_full=> fwr_sel_out_full, ack_out=> fwr_ack_out,
      dout=> mdwr, bout=> fwr_bout);
      
    fifo_rd: opb_master_fifo
    generic map(fifo_awidth=> fifo_rd_awidth1, awidth=> 1, use_be=> 0, Kin=> 1,
      dout_width=> mdwidth, din_width=> mdwidth,
      use_seqAddr=> use_seqAddr, transfer_size=> transfer_size)
    port map(clk=> clk, arst=> arst, rst=> rst,
      transfer_full=> frd_transfer_full, transfer_empty=> open,
      almost_full=> frd_almost_full, almost_empty=>frd_almost_empty, full=> frd_full,
      -- data input signals (slave)
      sel_in=> frd_sel_in, ack_in=> open, adr_in=> one,
      din=> mdrd, bin=> bin_rd_ones,
      -- data output signals (master)
      sel_out=> open, sel_out_full=> frd_sel_out_full, ack_out=> frd_ack_out,
      dout=> fdrd, bout=> open);
    
    empty<= '1' when fifo_rd_awidth<=0 and fifo_wr_awidth<= 0 else -- direct connection
      not frd_sel_out_full when reading='1' else 
      not fwr_sel_out when use_be_wr>0 else not fwr_sel_out_full; -- when use_be<=0 transfer only full data (disregard partial write)
    full<= frd_full when reading='1' else fwr_full;
    almost_full<= frd_almost_full when reading='1' else fwr_almost_full;
    almost_empty<= frd_almost_empty when reading='1' else fwr_almost_empty;
    transfer_full<= frd_transfer_full when wr_pipeline_latency<=0 or rnw='1' else fwr_transfer_full; -- transfer FULL is required only reading or for writes when wr_pipeline_latency>0
    ack<= fwr_ack_in when reading='0' else sel and frd_sel_out_full;
    msel1<= not frd_full and not mseqAddr_QN when reading='1' and (fifo_wr_awidth>0 or fifo_rd_awidth>0) else
      sel when reading='1' else
    fwr_sel_out_full when finish='0' else fwr_sel_out;
    fwr_sel_in<= sel and not reading;
    fwr_ack_out<= mack and not reading;
    frd_sel_in<= mack and reading;
    frd_ack_out<= sel when frd_sel_out_full='1' and 
    adr(lsb_mawidth to lsb_awidth-1)=(2**be_awidth)-1 else '0'; -- the last read
    mbe1<=  (others=>'1') when reading='1' or use_be_wr<=0 else fwr_bout;
    -- read data multiplexer
    mux: opb_master_mux
    generic map(din_width=> mdwidth, dout_width=> dwidth,
      sel_width=> be_awidth, K=> K)
    port map(sel=> adr(lsb_mawidth to lsb_awidth-1), 
      din=> fdrd, dout=> drd);
    --    process(adr(lsb_mawidth to lsb_awidth-1), fdrd) begin 
    --      for i in 0 to K-1 loop
    --        if adr(lsb_mawidth to lsb_awidth-1)=(2**be_awidth)-K + i then
    --          drd(0 to MIN((i+1)*dwidth, mdwidth)-i*dwidth-1)<= fdrd(i*dwidth to MIN((i+1)*dwidth, mdwidth)-1);
    --          if (i+1)*dwidth > mdwidth then
    --            drd((i+1)*dwidth-mdwidth to dwidth-1)<= (others=>'0');
    --          end if;
    --        end if;
    --      end loop;              
    --    end process; 
  end generate;
  
  
  eq1: if dwidth = mdwidth and (fifo_wr_awidth>0 or fifo_rd_awidth>0) generate -- no data width convertion - use only single FIFO for write and read logic
    fifo: opb_master_fifo -- only a single FIFO is used for both read and write operation
    generic map(fifo_awidth=> MAX(fifo_wr_awidth1, fifo_rd_awidth1),
      awidth=> 1, use_be=> use_be_wr, Kin=> 1,
      dout_width=> dwidth, din_width=> dwidth,
      use_seqAddr=> use_seqAddr, transfer_size=> transfer_size + wr_pipeline_latency)
    port map(clk=> clk, arst=> arst, rst=> rst,
      transfer_full=> transfer_full, transfer_empty=> transfer_empty,
      almost_full=> almost_full, almost_empty=>almost_empty, full=> full,
      -- data input signals (slave)
      sel_in=> fwr_sel_in, ack_in=> fwr_ack_in, 
      adr_in=> one, din=> fdin, bin=> be,
      -- data output signals (master)
      sel_out=> open, sel_out_full=> fwr_sel_out_full, ack_out=> fwr_ack_out,
      dout=> fdout, bout=> fwr_bout);
    
    empty<= not fwr_sel_out_full; 
    fdin<= mdrd when reading='1' or fifo_wr_awidth<0 else -- FIFO input select (multiplexer)
    dwr when fifo_wr_awidth>0 or fifo_rd_awidth>0 else (others=>'0'); -- use FIFO at all
    drd<= (others=>'0') when fifo_rd_awidth<0 else -- no read logic
    fdout; -- use fifo
    mdwr<= (others=>'0') when fifo_wr_awidth<0 else -- do not use write logic at all
    fdout; -- FIFO output
    fwr_sel_in<= mack when reading='1' else sel;
    fwr_ack_out<= sel and fwr_sel_out_full when reading='1' else mack when fifo_wr_awidth>=0 else '0';
    ack<= sel and fwr_sel_out_full when reading='1' else fwr_ack_in;
    msel1<= fwr_sel_out_full when reading='0' else not full and not mseqAddr_QN;
    mbe1<=  (others=>'1') when reading='1' or use_be<=0 else fwr_bout;
  end generate; -- K=1 and fifo_awidth>1
  
  eq0: if dwidth = mdwidth and fifo_wr_awidth<=0 and fifo_rd_awidth<=0 generate -- direct connection
    transfer_full<= '1';
    transfer_empty<= '1';
    almost_full<= '1';
    almost_empty<= '1';
    empty<= '1';
    drd<= (others=>'0') when fifo_rd_awidth<0 else mdrd;
    mdwr<= (others=>'0') when fifo_wr_awidth<0 else dwr;
    ack<= mack;
    msel1<= sel;
    mbe1<= (others=>'1') when use_be<=0 else be;
  end generate;
  
  mbe<= mbe1;
  madr_out(0 to lsb_mawidth-1)<= madr_in(0 to lsb_mawidth-1);
  gadr0: if mdwidth>8 and use_be_wr<=0 generate
    madr_out(lsb_mawidth to mawidth-1)<= (others=>'0');
  end generate;
  gadr1: if use_be_wr>0 and mdwidth>8 and mdwidth<=16 generate
    madr_out(mawidth-1)<= '0' when mbe1(0)='1' else '1';
  end generate;
  gadr2: if use_be_wr>0 and mdwidth>16 and mdwidth<=32 generate
    madr_out(mawidth-2 to mawidth-1)<= "00" when mbe1(0)='1' else 
    "01" when mbe1(1)='1' else
    "10" when mbe1(2)='1' else "11";
  end generate; 
  
end arch;



----------------------------------------------------------------------------------------------------------------------
-- opb_master_aux -- use opb_master_fifo and additional logic to generate proper logic for master
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;
use IEEE.std_logic_misc.all;

entity opb_master_aux is
  generic(awidth: integer:= 16; -- address width (user side)
    mawidth: integer; -- master address width
    opb_awidth: integer; -- same as c_opb_awidth
    dwidth: integer:= 8; -- data width
    real_dwidth: integer; -- opb bus dwidth
    use_BE: integer:= 1; -- =1 - use Byte Enable logic or =0 - assume that be is always '1'
    fifo_wr_awidth: integer:= 0; 
    fifo_rd_awidth: integer:= 0;
    use_seqAddr: integer:= 1; -- use seqAddressing
    single_master: integer; 
    user_full_transfer: integer; 
    transfer_size: integer:= 4; -- same as for the opb_master
    use_addr_counter: integer; -- use internal address counter
    wr_pipeline_latency: integer);
  port(clk, arst: in std_logic; --clock and global asynchronous reset, and synch reset
    -- user logic
    adr: in std_logic_vector(0 to awidth-1);
    dwr: in std_logic_vector(0 to dwidth-1);
    drd: out std_logic_vector(0 to dwidth-1);
    be: in std_logic_vector(0 to (dwidth+7)/8-1);
    sel: in std_logic;
    rnw: in std_logic;
    seqAddr: in std_logic;
    ack: out std_logic;
    fifo_full, fifo_empty: out std_logic; -- fifo is full and empty
    fifo_almost_full, fifo_almost_empty: out std_logic; -- only one data can be transferred and fifo will be full or empty
    retry: out std_logic; -- FIFO state indicates that it is better to dwaw retry signal (this signal might be used by an opb_slave module connected to this opb_master module)
    load_addr: in std_logic; -- load internal address counter
    -- master logic - to connect to the master
    mrst: in std_logic; -- synchronous reset
    mbus_granted: in std_logic; -- OPB bus is not granted
    mbus_req: out std_logic; -- OPB bus request logic - optimise for sequencial transfers
    madr: out std_logic_vector(0 to mawidth-1);
    mdwr: out std_logic_vector(0 to real_dwidth-1);
    mdrd: in std_logic_vector(0 to real_dwidth-1);
    mbe: out std_logic_vector(0 to (real_dwidth+7)/8-1);
    msel: out std_logic;
    mrnw: out std_logic;
    mseqAddr: out std_logic;
    mack: in std_logic);
end opb_master_aux;               

architecture arch of opb_master_aux is
  component opb_master_fifo2
    generic(awidth: integer:= 4; -- user address bus width - LSBs of the address bus are taken into account
    mawidth: integer:= 4; -- master side address width - the same as awidth for dwidth>4
    dwidth: integer:= 8; -- user data width
    mdwidth: integer:= 16; -- opb Bus data width
    use_be: integer:= 1; -- use byte enable
    use_seqAddr: integer:= 1; -- use sequantiaal addressing
    user_full_transfer: integer; 
    fifo_wr_awidth: integer:= 4; -- write fifo address width (rnw=0)
    fifo_rd_awidth: integer:= 4; -- read fifo address width
    wr_pipeline_latency: integer:= 0; 
    transfer_size: integer:= 4); -- the transfer size (same as opb_master)
  port (clk: in std_logic; -- clock
    arst: in std_logic; -- global asynchronous reset
    rst: in std_logic; -- synchronous reset for the whole device
    rnw: in std_logic; -- read not write operation
    transfer_full, transfer_empty: out std_logic; -- transfer_full = 1 - fifo may not accept transfer_size data, transfer_empty - fifo contains less than transfer_size data
    almost_full, almost_empty: out std_logic; -- only one data can be accepted (transfered in or out)
    full: buffer std_logic;
    empty: out std_logic; -- empty (no data in the buffer)
    finish: in std_logic; -- finish - generate msel for write even when not full output word is valid
    -- user interface
    adr: in std_logic_vector(0 to awidth-1);
    dwr: in std_logic_vector(0 to dwidth-1);
    drd: out std_logic_vector(0 to dwidth-1);
    sel: in std_logic;
    seqAddr: in std_logic;
    ack: out std_logic;
    be: in std_logic_vector(0 to (dwidth+7)/8-1);
    -- data output signals (master)
    mseqAddr_in: in std_logic; -- seqAddressing - generated in opb_master_aux
    madr_out: out std_logic_vector(0 to mawidth-1);
    madr_in: in std_logic_vector(0 to mawidth-1);
    mdwr: out std_logic_vector(0 to mdwidth-1);
    mdrd: in std_logic_vector(0 to mdwidth-1);
    mbe: out std_logic_vector(0 to (mdwidth+7)/8-1);
    msel: out std_logic;
    mack: in std_logic);
  end  component;
  function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then 
      return arg0;
    else 
      return arg1;
    end if;
  end MAX;
  function MIN(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0<=arg1 then 
      return arg0;
    else 
      return arg1;
    end if;
  end MIN;
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
    for i in 0 to 32 loop
      if 2**i >= arg then ret:= i; exit; end if;
    end loop;
    return ret;
  end log2;
  function FindPipeDWidth(ignore: integer) return integer is begin -- Find the width of the pipeline buffer (sel signal and Byte Enable)
    if use_be<=0 or dwidth<=8 then
      return 1;
    else
      return 1 + (dwidth+7)/8;
    end if;
  end FindPipeDWidth;
  constant pipe_dwidth: integer:= FindPipeDWidth(0); -- pipline buffer data width - for sel signal and byte enable
  
  constant log2bit: integer:= log2(8/real_dwidth) - log2(8/dwidth); -- additional address length caused by data width <=4
  constant min_awidth: integer:= MIN(awidth, mawidth); 
  constant log2be: integer:= log2((real_dwidth+7)/8); -- master be
  constant log2dwidth: integer:= log2((dwidth+7)/8);
  constant log2fifo_adr: integer:= log2((real_dwidth+user_full_transfer*(dwidth-1))/dwidth)+ log2dwidth; -- address width that must be generated locally
  constant wr_pipeline_latency1: integer:= MAX(1, wr_pipeline_latency); --
  constant be_width: integer:= (dwidth+7)/8; -- input byte enable width
  signal pipe_sel: std_logic_vector(0 to wr_pipeline_latency1-1); -- pipeline select signals
  signal pipe_be: std_logic_vector(0 to wr_pipeline_latency1*be_width-1); -- pipeline byte enable
  signal pipe_or: std_logic; -- OR (pipeline)
  signal fifo_adr: std_logic_vector(0 to awidth-1); -- fifo input address (user side)
  signal madr1: std_logic_vector(0 to mawidth-1); -- internal (before fifo) madr
  signal adr_seq: std_logic_vector(0 to awidth-log2be-1); -- internal address logic
  signal adr_q: std_logic_vector(0 to mawidth-log2be-1); -- internal address logic
  signal valid_adr_rd, sel1, adr_last: std_logic; -- valid read address (or read operation is performed)
  signal fifo_sel, msel1, ack1, fifo_ack, seqAddr_q, seqAddr1, seqAddr2: std_logic;
  signal fifo_almost_full1, fifo_full1, fifo_transfer_full, fifo_rst: std_logic;
  signal mseqAddr1, mseqAddrQ, fifo_almost_empty1, fifo_empty1, fifo_transfer_empty, fifo_finish: std_logic;
begin
  -- pipeline logic
  gpipe1: if wr_pipeline_latency>0 generate
    process(arst, clk) begin -- pipeline shift register
      if arst='1' then pipe_sel<= (others=>'0');
      elsif clk'event and clk='1' then
        if mrst='1' then
          pipe_sel<= (others=>'0');
        else
          if wr_pipeline_latency>1 then
            pipe_sel(0 to wr_pipeline_latency-2)<= pipe_sel(1 to wr_pipeline_latency-1); -- shift
          end if;
          if ack1='1' and valid_adr_rd='0' then
            pipe_sel(wr_pipeline_latency-1)<= '1';
          else pipe_sel(wr_pipeline_latency-1)<= '0';
          end if;
        end if;
        
        if use_be>0 and dwidth>8 then -- byte enable shift register
          if wr_pipeline_latency>1 then
            pipe_be(0 to (wr_pipeline_latency-1)*be_width-1)<= pipe_be(be_width to wr_pipeline_latency*be_width-1) & be;
          else 
            pipe_be(0 to be_width-1)<= be;
          end if;
        end if;
      end if;
    end process;
    pipe_or<= OR_REDUCE(pipe_sel);
  end generate; -- wr_pipeline_latency>0
  gpipe0: if wr_pipeline_latency<=0 generate
    pipe_or<= '0';
    pipe_sel(0)<= '0';
    pipe_be<= be;
  end generate;
  
  -- internal input (fifo) LSB address
  ga0: if dwidth>=real_dwidth or (wr_pipeline_latency<=0 and use_addr_counter<=0) generate
    fifo_adr<= adr;
  end generate;
  ga1: if dwidth < real_dwidth and (wr_pipeline_latency>0 or use_addr_counter>0) generate -- generate internal address for fifo only when fifo_adr is different from input address and dwidth<real_dwidth
    --fifo_adr(0 to awidth-log2fifo_adr-1)<= (others=>'0');
    process(arst, clk) begin -- internal address counter only for address width used by fifo
      if arst='1' then fifo_adr(awidth-log2fifo_adr to awidth-log2dwidth-1)<= (others=>'0');
      elsif clk'event and clk='1' then
        if (use_addr_counter>0 and load_addr='1') or
          (use_addr_counter<=0 and ack1='1' and seqAddr_Q='0') then
          fifo_adr(awidth-log2fifo_adr to awidth-log2dwidth-1)<= adr(awidth-log2fifo_adr to awidth-log2dwidth-1);
        elsif fifo_ack='1' then
          fifo_adr(awidth-log2fifo_adr to awidth-log2dwidth-1)<= fifo_adr(awidth-log2fifo_adr to awidth-log2dwidth-1)+1;
        end if;
      end if;
    end process;
  end generate;  -- internal address counter for FIFO LSB address
  
  
  -- detect seqAddressing based on the address bus
  gseqAddr2: if use_seqAddr=2 and use_addr_counter<=0 generate
    process(arst, clk) begin
      if arst='1' then adr_seq<= (others=>'0'); seqAddr2<= '0';
      elsif clk'event and clk='1' then
        seqAddr2<= seqAddr1;
        if ack1='1' or mrst='1' then
          adr_seq<= adr(0 to awidth-1-log2be);
        end if;
      end if;
    end process;
    seqAddr1<= '1' when sel='0' or seqAddr2='0' else -- ignore adr when sel=0 and assume that the new transfer is seq.
      '1' when adr(0 to awidth-1-log2be)-adr_seq-1=0 else '0';
    sel1<= sel and (seqAddr1 or not seqAddr2); 
  end generate;
  
  -- use seqAddr signal
  gseqAddr1: if use_seqAddr/=2 or use_addr_counter>0 generate
    sel1<= sel;
    seqAddr1<= '0' when use_seqAddr<=0 else seqAddr;
  end generate; 
  
  
  fifo: opb_master_fifo2
  generic map(awidth=> awidth, mawidth=> mawidth, dwidth=> dwidth, mdwidth=> real_dwidth,
    use_be=> use_be, use_seqAddr=> use_seqAddr, user_full_transfer=> user_full_transfer,
    fifo_wr_awidth=> fifo_wr_awidth, fifo_rd_awidth=> fifo_rd_awidth,
    wr_pipeline_latency=> wr_pipeline_latency, transfer_size=> transfer_size)
  port map (clk=> clk, arst=> arst, rst=> fifo_rst, rnw=> valid_adr_rd,
    transfer_full=> fifo_transfer_full, transfer_empty=> fifo_transfer_empty,
    almost_full=> fifo_almost_full1, almost_empty=> fifo_almost_empty1,
    full=> fifo_full1, empty=> fifo_empty1, finish=> fifo_finish,
    -- user interface
    adr=> fifo_adr, dwr=> dwr, drd=> drd, seqAddr=> seqAddr,
    sel=> fifo_sel, ack=> fifo_ack, be=> pipe_be(0 to be_width-1),
    -- data output signals (master)
    madr_out=> madr, madr_in=> madr1, 
    mdwr=> mdwr, mdrd=> mdrd, mbe=> mbe, mseqAddr_in=> mseqAddr1,
    msel=> msel1, mack=> mack);
  
  fifo_sel<= '1' when pipe_sel(0)='1' and wr_pipeline_latency>0 else -- write from pipeline logic
    '0' when wr_pipeline_latency>0 and (valid_adr_rd='0' or fifo_rd_awidth<0) else -- no write with pipeline logic
    '0' when rnw='0' and (fifo_empty1='0' or pipe_or='1') and (seqAddr_Q='0' or use_seqAddr<=0) else -- valid data from previous writes or reads
    '0' when fifo_rd_awidth>=0 and rnw='1' and valid_adr_rd='0' else -- address is not valid for reading
    sel1;
  
  fifo_full<= fifo_full1;
  fifo_almost_full<= fifo_almost_full1;
  fifo_almost_empty<= fifo_almost_empty1;
  fifo_empty<= fifo_empty1;
  --fifo_rst<= mrst or (valid_adr_rd and not seqAddr1 and (not sel1 or fifo_ack)); -- reset when external reset or not seqAddr and the end of the read transfer
  ack1<= fifo_ack when wr_pipeline_latency<=0 or (rnw='1' and fifo_rd_awidth>=0) else -- no pipeline or read logic
  sel1 when (fifo_empty1='1' and pipe_or='0') or -- the first transfer
  (seqAddr_Q='1' and use_seqAddr>0 and fifo_transfer_full='0') else '0';  -- next seqAddr transfers
  ack<= ack1;
  msel<= msel1;
  
  
  g1: if fifo_wr_awidth>=1 or fifo_rd_awidth>=1 generate -- insert fifo
    -- seqAddressing - detect SeqAddr low signal
    gs1: if use_seqAddr>0 generate
      process(arst, clk) begin
        if arst='1' then seqAddr_q<= '0';
        elsif clk'event and clk='1' then
          if use_addr_counter>0 then
            seqAddr_q<= not load_addr and seqAddr1;
          elsif seqAddr1='0' and (sel1='0' or fifo_ack='1') then -- end of seqAddr read
            seqAddr_q<= '0'; -- detect seqAddr low
          elsif fifo_empty1='1' and sel1='1' then -- fifo is empty
            seqAddr_q<= seqAddr1;
          end if;
        end if;
      end process;
    end generate;
    -- address counter  
    process(arst, clk) begin
      if arst='1' then adr_q<= (others=>'0'); adr_last<= '0';
      elsif clk'event and clk='1' then
        if mrst='1' and use_addr_counter>0 and (use_addr_counter=2 or fifo_wr_awidth>=0) then -- reset logic should be implemented only when internal address counter is used and load signal is not required
          adr_q<= (others=>'0'); adr_last<= '0';
        elsif (load_addr='1' and use_addr_counter>0) or -- internal address counter
          (use_addr_counter<=0 and sel1='1' and fifo_empty1='1' and (seqAddr_Q='0' or use_seqAddr=0)) then -- the first transfer
          adr_q(0 to min_awidth-log2be-1)<= adr(0 to min_awidth-log2be-1);
          if log2bit>0 then -- output data width <=4 so the extra LSB(s) address bits should be initialised with zeros
            adr_q(mawidth-log2bit to mawidth-1)<= (others=>'0');
          end if;
          adr_last<= '0';
        elsif (use_seqAddr>0 or use_addr_counter>0) and mack='1' then
          adr_q<= adr_q + 1; -- increase the address locally
          if adr_q= 2**(awidth-log2be)-2 and awidth<=24 and opb_awidth>awidth then -- next to the last address, - do not use address_last logic for address width >=24 address overflow is rather not possible - do not use this logic when opb_awidth<=awidth
            adr_last<='1';
          else adr_last<= '0';
          end if;
        end if;
      end if;
    end process; 
    
    
    madr1(0 to mawidth-log2be-1)<= adr_q;
    gadr1: if log2be>=1 generate
      madr1(mawidth-log2be to mawidth-1)<= (others=>'0');
    end generate;
    
    mrnw<= '1' when fifo_wr_awidth<0 else
      '0' when fifo_rd_awidth<0 else
      valid_adr_rd;
    mseqAddr1<= '0' when adr_last='1' or use_seqAddr<=0 or mseqAddrQ='0' else -- last address location
      '1' when single_master>0 and seqAddr_Q='1' else
      not fifo_almost_empty1 when fifo_wr_awidth>=0 and (valid_adr_rd='0' or fifo_rd_awidth<0) else -- write operation
      '0' when fifo_rd_awidth<1 else -- do not use seqAddressing for reading
    seqAddr_Q and not fifo_almost_full1; -- reading
    mseqAddr<= mseqAddr1;
    
    -- mseqAddrQ logic - the state of this signal cannot change when msel='1' and xferAck='0' because the state of seqAddr cannot change from 0 to 1 at the same tranfer cycle.
    mseqAddrQ1: if use_seqAddr>0 generate
      process(arst, clk) begin
        if arst='1' then mseqAddrQ<= '0';
        elsif clk'event and clk='1' then
          if msel1='0' or mbus_granted='0' or mack='1' or mseqAddrQ='1' then -- can change the state
            if fifo_wr_awidth>=0 and (valid_adr_rd='0' or fifo_rd_awidth<0) then -- write operation
              mseqAddrQ<= not fifo_almost_empty1;
            else -- read operation
              mseqAddrQ<= seqAddr_Q and not fifo_almost_full1;
            end if;
          end if;
        end if;
      end process;
    end generate;   
    
    fifo_rst<= '1' when mrst='1' else -- external reset
      load_addr when use_addr_counter>0 else -- load address so reset the FIFO
      '1' when valid_adr_rd='1' and (use_seqAddr<=0 or seqAddr1='0') and (sel1= '0' or msel1='1') else -- read operation finished
      '0';
    fifo_finish<= '1' when use_seqAddr<=0 else 
      '0' when use_be<=0 else -- fr use_be<=0 use only full transfers (so do not write a partial result)
      not seqAddr_Q; -- finish the write operation when no more data to be written in sequence
    retry<= '1' when fifo_empty1='0' and (single_master>0 or mbus_granted='0' or fifo_full1='1') and -- fifo contains valid write data and bus is not granted
      (seqAddr_Q='0' or use_seqAddr<=0 or fifo_full1='1') and valid_adr_rd='0' else '0';
    
    gval_adr_log: if fifo_rd_awidth>=0 and use_addr_counter<=0 generate  -- valid_adr for read operation
      process(arst, clk) begin
        if arst='1' then valid_adr_rd<= '0';
        elsif clk'event and clk='1' then
          if mrst='1' or ((use_seqAddr<=0 or seqAddr1='0') and (sel1= '0' or fifo_empty1='0')) then -- the last read or do not use seqAddressing
            valid_adr_rd<= '0';
          elsif sel1='1' and rnw='1' and fifo_empty1='1' and pipe_or='0' then
            valid_adr_rd<= '1';
          end if;
        end if;
      end process;
    end generate;
    
  end generate;  -- use fifo_awidth>0
  
  
  -- direct and indirect logic
  -- use address counter (only address counter logic)  -- wait for the first load_addr='1'
  g_val_adr1: if  fifo_rd_awidth>=0 and use_addr_counter>0 and 
    (use_addr_counter/=2 or fifo_wr_awidth>=0)  generate
    process(arst, clk) begin
      if arst='1' then valid_adr_rd<= '0';
      elsif clk'event and clk='1' then
        if mrst='1' then
          valid_adr_rd<='0'; 
        elsif load_addr='1' and (rnw='1' or fifo_wr_awidth<0) then
          valid_adr_rd<= '1';
        end if;
      end if;
    end process;
  end generate; 
  
  -- start reading from zero (do not wait for load_addr)
  g_val_adr2: if fifo_rd_awidth>=0 and use_addr_counter=2 and fifo_wr_awidth<0 generate
    valid_adr_rd<= '1';
  end generate;
  
  -- reading is not implemented at all 
  g_val_adr0: if fifo_rd_awidth<0 generate
    valid_adr_rd<= '0';
  end generate;
  
  
  -- direct logic
  g_direct: if fifo_wr_awidth<=0 and fifo_rd_awidth<=0 generate 
    g_val_adr3: if use_addr_counter<=0 generate
      valid_adr_rd<= '1' when fifo_wr_awidth<0 else rnw when fifo_rd_awidth>=0 else '0';
    end generate;
    mseqAddr<= '0' when use_seqAddr<=0 or adr_last='1' else
    '0' when load_addr='1' and use_addr_counter>0 else
    seqAddr1;
    mrnw<= rnw when fifo_rd_awidth=0 and fifo_wr_awidth=0 else '1' when fifo_rd_awidth=0 else '0';
    fifo_finish<= '1'; -- finish the calculation process
    retry<= '0';
    ga0: if use_addr_counter<= 0 generate
      madr1(0 to min_awidth-1)<= adr(0 to min_awidth-1);
      adr_last<= '1' when adr(0 to awidth-log2be-1)= 2**(awidth-log2be)-1 and awidth<=24 and awidth<opb_awidth else '0'; -- next to the last address, - do not use address_last logic for address width >=24 address overflow is rather not possible
    end generate;
    -- use internal address counter
    ga1: if use_addr_counter>0 generate
      process (clk, arst) begin
        if arst='1' then adr_q<= (others=>'0');
        elsif clk'event and clk='1' then
          if mrst='1' and (use_addr_counter=2 or fifo_wr_awidth>=0) then
            adr_q<= (others=>'0');
          elsif load_addr='1' then
            adr_q(0 to min_awidth-1-log2be)<= adr(0 to min_awidth-1-log2be);
            if mawidth>awidth then
              adr_q(awidth to mawidth-1)<= (others=>'0');
            end if;
          elsif mack='1' then
            adr_q<= adr_q + 1;
          end if;
        end if;
      end process;
      madr1(0 to mawidth-1-log2be)<= adr_q;
      gadr_lsb: if log2be> 0 generate
        madr1(mawidth-log2be to mawidth-1)<= (others=>'0');
      end generate;
      adr_last<= '1' when adr_q(0 to mawidth-log2be-1)= 2**(mawidth-log2be)-1 and awidth<=24 and awidth<opb_awidth else '0';
    end generate;
  end generate;
  
  mbus_req<= '0' when (fifo_wr_awidth<0 and fifo_rd_awidth<0) else -- master logic is not valid
  sel1 and not mbus_granted when fifo_wr_awidth<=0 and fifo_rd_awidth<=0 else -- direct connection
  -- read operation
  fifo_empty1 and not mbus_granted  when valid_adr_rd='1' and  fifo_rd_awidth>=0 and 
  (use_seqAddr=0 or seqAddr_Q='0' or fifo_rd_awidth<=1 or transfer_size>=2**fifo_rd_awidth) else -- not seqAddressing so address is valid and read fifo is empty
  not fifo_transfer_full  when valid_adr_rd='1' and fifo_rd_awidth>1 else -- sequential addresssing so try to ransfer new data only when transfer_size data can be accepted
  -- write operation
  '0' when fifo_wr_awidth<0 or msel1='0' else -- no write logic
  not fifo_almost_empty1 or not mbus_granted -- do not generate bus_request when only a single data are in the FIFO and bus is granted
  when use_seqAddr=0 or fifo_wr_awidth<=1 or seqAddr_Q='0' or transfer_size>=2**fifo_wr_awidth or use_addr_counter=1 else -- no seqAddr or the end of seqAddressing
  not fifo_transfer_empty; -- write fifo contains at least transfer_size data
  
  
  --  assert not (wr_pipeline_latency>0 and transfer_size + wr_pipeline_latency >= 2**fifo_wr_awidth)
  --    report "opb_master, it is not recommended that transfer_size + pipeline_latency >= 2**fifo_awidth"
  --    severity warning;
  --  assert not (transfer_size> 2**fifo_wr_awidth and dwidth<=real_dwidth and fifo_wr_awidth>0)
  --    report "opb_master, it is not recommended that transfer_size>2**fifo_wr_awidth"
  --    severity warning;
  --  assert not (transfer_size> 2**fifo_rd_awidth and dwidth>=real_dwidth and fifo_rd_awidth>0)
  --    report "opb_master, it is not recommended that transfer_size>2**fifo_rd_awidth"
  --    severity warning;
  --  assert not (use_be>0 and dwidth /= real_dwidth and use_seqAddr<=0)
  --    report "opb_master(slave), use_seqAddr<=0 and data width conversion therefore a single dwidth-bit transfer is translated only to a single real_dwidth-bit transfer"
  --    severity warning;
  --  assert not (use_be>0 and dwidth /= real_dwidth and (dwidth rem 8)/= 0)
  --    report "opb_master(slave), dwidth are not multiple of 8 therefore Byte Enable logic cannot be implemented (the parameter use_be is internally set to 0)"
  --    severity warning;
  --  assert not (wr_pipeline_latency>0 and wr_pipeline_latency>= 2**fifo_wr_awidth)
  --    report "opb_master, the write FIFO address width is too small for the defined wr_pipeline_latency  and transfer_size parameter"
  --    severity error;    
  
  --  assert not(clk'event and clk='1' and use_be<=0 and dwidth>8 and 
  --     use_addr_counter<=0 and sel='1' and rnw='0' and adr/=(awidth-log2dwidth to awidth-1 => '0'))
  --    report "opb_master. UseBe<=0 (Byte Enable) and address bus indicates that not full data bus write is used, result: data corruption"
  --    severity warning;  
  
  --  process(clk) begin
  --    if clk'event and clk='1' then
  --      assert not (use_addr_counter>0 and fifo_empty1='0' and valid_adr_rd='0' and load_addr='1')
  --        report "opb_master. Address counter is loaded (load_addr='1') and write FIFO is not empty - this causes that write data are lost"
  --        severity warning;     
  --      assert not (fifo_wr_awidth<0 and sel1='1' and rnw='0')
  --        report "opb_master. Write operation is performed but write path is not implemented (fifo_wr_awidth<0)"
  --        severity error;
  --      assert not (fifo_rd_awidth<0 and sel1='1' and rnw='1')
  --        report "opb_master(slave). Read operation is performed but read path is not implemented (fifo_rd_awidth<0)"
  --        severity error;
  --     end if;
  --   end process;
end arch;



----------------------------------------------------------------------------------------------------------------------

architecture opb_master_arch of opb_master is
  component opb_master_aux
    generic(awidth: integer:= 16; -- user address width
      mawidth: integer:= 16; -- master address width
      opb_awidth: integer; -- same as c_opb_awidth
      dwidth: integer:= 8; -- data width
      real_dwidth: integer; -- opb bus dwidth
      use_BE: integer:= 1; -- =1 - use Byte Enable logic or =0 - assume that be is always '1'
      fifo_wr_awidth: integer:= 0; 
      fifo_rd_awidth: integer:= 0;
      use_seqAddr: integer:= 1; -- use seqAddressing
      single_master: integer; 
      user_full_transfer: integer; 
      transfer_size: integer:= 4; -- same as for the opb_master
      use_addr_counter: integer; -- use internal address counter
      wr_pipeline_latency: integer);
    port(clk, arst: in std_logic; --clock and global asynchronous reset
      -- user logic
      adr: in std_logic_vector(0 to awidth-1);
      dwr: in std_logic_vector(0 to dwidth-1);
      drd: out std_logic_vector(0 to dwidth-1);
      be: in std_logic_vector(0 to (dwidth+7)/8-1);
      sel: in std_logic;
      rnw: in std_logic;
      seqAddr: in std_logic;
      ack: out std_logic;
      fifo_full, fifo_empty: out std_logic; -- fifo is full and empty
      fifo_almost_full, fifo_almost_empty: out std_logic; -- only one data can be transferred and fifo will be full or empty
      retry: out std_logic; -- FIFO state indicates that it is better to dwaw retry signal (this signal might be used by an opb_slave module connected to this opb_master module)
      load_addr: in std_logic; -- load internal address counter
      -- master logic - to connect to the master
      mrst: in std_logic; -- synchronous reset
      mbus_granted: in std_logic; -- OPB bus is not granted
      mbus_req: out std_logic; -- OPB bus request logic - optimise for sequencial transfers
      madr: out std_logic_vector(0 to mawidth-1);
      mdwr: out std_logic_vector(0 to real_dwidth-1);
      mdrd: in std_logic_vector(0 to real_dwidth-1);
      mbe: out std_logic_vector(0 to (real_dwidth+7)/8-1);
      msel: out std_logic;
      mrnw: out std_logic;
      mseqAddr: out std_logic;
      mack: in std_logic);
  end component;
  
  function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
    for i in 0 to 32 loop
      if 2**i >= arg then ret:= i; exit; end if;
    end loop;
    return ret;
  end log2;
  function MAX(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0>=arg1 then 
      return arg0;
    else 
      return arg1;
    end if;
  end MAX;
  function MIN(arg0: integer; arg1: integer) return integer is
    variable ret: integer;
  begin
    if arg0<=arg1 then 
      return arg0;
    else 
      return arg1;
    end if;
  end MIN;
  
  function FindBusDwidth(ignore: integer) return integer is begin
    if real_dwidth<=0 then
      return dwidth;
    else
      return real_dwidth;
    end if;
  end FindBusDwidth;
  constant real_dwidth1: integer:= FindBusDwidth(0);
  
  -- change the fifo_awidth to 1 when use_be<=0 and different data width and not seqAddressing
  function FindFifoAwidth(fifo_awidth: integer) return integer is begin 
    if use_seqAddr<=0 and use_be<=0 and real_dwidth1/=dwidth and fifo_awidth=0 then
      return 1;
    else 
      return fifo_awidth;
    end if;
  end FindFifoAwidth;
  constant fifo_wr_awidth1: integer:= FindFifoAwidth(fifo_wr_awidth);
  constant fifo_rd_awidth1: integer:= FindFifoAwidth(fifo_rd_awidth);
  constant use_seqAddr1: integer:= FindFifoAwidth(use_seqAddr); -- use_seqAddr=> 1 when data width conversion
  constant mawidth: integer:= awidth + log2(8/real_dwidth1) - log2(8/dwidth);
  
  -- bus arbitration signals
  constant transfer_size1: integer:= MAX(1, transfer_size);
  signal bus_granted: std_logic; -- OPB bus is granted
  signal OPB_sel: std_logic; -- internal OPB select signal
  signal transfer_count: integer range 0 to transfer_size1-1; -- how many transfer since bus master
  signal pendReq_q: std_logic; -- another device also request the bus - flip-flop output
  signal seqAddr1: std_logic; -- similar like seqAddr but OPB logic is added
  signal transfer_last: std_logic; -- last transfer
  -- fifo and internal adres logic
  signal a_adr: std_logic_vector(0 to mawidth-1);
  signal a_dwr, a_drd: std_logic_vector(0 to real_dwidth1-1);
  signal a_be: std_logic_vector(0 to (real_dwidth1+7)/8-1);
  signal a_sel, a_ack, a_rnw, a_seqAddr, a_bus_req, a_retry: std_logic;
  
begin
  g1: if single_master<=0 generate 
    -- bus_granted logic
    pbus_grant: process(clk, arst) begin
      if arst='1' then bus_granted<= '0';
      elsif clk'event and clk='1' then
        if OPB_MGrant='1' then
          bus_granted<= '1';
        elsif OPB_retry='1' or OPB_timeout='1' or OPB_errAck='1' -- error condition
          or a_sel='0' -- device is not ready for data transfer 
          or (OPB_xferAck='1' and (a_seqAddr='0' -- do not use sequencial addressing or not seqAddr cycle
          or (transfer_last='1' and pendReq_q='1'))) then -- relinquesh bus only when another master requensts the bus
          bus_granted<= '0';
        end if;
      end if;
    end process;								   
    
    transfer_last<= '1' when transfer_count= transfer_size1-1 else '0';
    
    ptransfer_count: process(clk, arst) begin
      if arst='1' then transfer_count<= 0; pendReq_q<= '0';
      elsif clk'event and clk='1' then
        pendReq_q<= OPB_pendReq or (pendReq_q and bus_granted); -- keep pendReq_q high until the bus is not granted - result of the OPB_bus arbiter error
        if bus_granted='0' then
          transfer_count<= 0;
        elsif transfer_last='0' and OPB_xferAck='1' then
          transfer_count<= transfer_count + 1; -- counter with saturation
        end if;
      end if;
    end process;		
  end generate; -- more than one master
  
  g0: if single_master>0 generate
    bus_granted<= '1';
    pendReq_q<= '0';
  end generate;
  
  OPB_sel<= bus_granted and a_sel;
  M_select<= OPB_sel;
  
  gd: if real_dwidth1<opb_dwidth generate
    M_BE((real_dwidth1+7)/8 to opb_dwidth/8-1)<= (others=>'0');
    M_DBus(real_dwidth1 to opb_dwidth-1)<= (others=>'0'); -- the LSBs are always zero
  end generate;
  
  M_BE(0 to (real_dwidth1+7)/8 -1)<= a_be when bus_granted='1' else (others=>'0');
  M_DBus(0 to real_dwidth1-1)<= a_dwr when bus_granted='1' and a_rnw='0' else (others=>'0');
  M_RNW<= a_rnw when bus_granted='1' else '0';
  seqAddr1<= '0' when bus_granted='0' else
  '0' when transfer_last='1' and pendReq_q='1' and single_master<=0 else -- when the last cycle
  a_seqAddr when a_sel='1' or pendReq_q='0' or single_master>0 else '0';
  M_seqAddr<= seqAddr1;
  M_busLock<= seqAddr1;
  
  M_request<= '1' when single_master>0 else
  '0' when a_bus_req='0' else
  '1' when bus_granted='0' else
  '1' when pendReq_q='0' else '0' ;
  a_ack<= OPB_xferAck and bus_granted and a_sel; -- sometimes ack is genereted even if the sel is not longer active
  a_drd<= OPB_DBus(0 to real_dwidth1-1);
  
  M_ABus(opb_awidth-mawidth to opb_awidth-1)<= a_adr when bus_granted='1' else (others=>'0');  
  ga: if opb_awidth > mawidth generate
    M_ABus(0 to opb_awidth-mawidth-1)<= msb_adr(0 to opb_awidth-mawidth-1) 
    when bus_granted='1' else (others=>'0');
  end generate;
  
  retry<= a_retry when fifo_wr_awidth>0 or fifo_rd_awidth>0 else OPB_retry;
  
  aux: opb_master_aux
  generic map(awidth=> awidth, mawidth=> mawidth, opb_awidth=> opb_awidth,
    dwidth=> dwidth, real_dwidth=> real_dwidth1, user_full_transfer=> MAX(0, MIN(1, user_full_transfer)),
    use_BE=> use_BE, single_master=> single_master,
    fifo_wr_awidth=> fifo_wr_awidth1, fifo_rd_awidth=> fifo_rd_awidth1, use_addr_counter=> use_addr_counter,
    use_seqAddr=> use_seqAddr1, transfer_size=> transfer_size, wr_pipeline_latency=> wr_pipeline_latency)
  port map(clk=> clk, arst=> arst,
    adr=> a, dwr=> di, drd=> do, be=> be, sel=> sel, rnw=> rnw, seqAddr=> seqAddr, ack=> ack, 
    load_addr=> load_addr, fifo_full=> fifo_full, fifo_empty=> fifo_empty, retry=> a_retry,
    fifo_almost_empty=> fifo_almost_empty, fifo_almost_full=> fifo_almost_full,
    -- master logic - to connect to the master
    mrst=> OPB_rst, mbus_req=> a_bus_req, mbus_granted=> bus_granted,
    madr=> a_adr, mdwr=> a_dwr, mdrd=> a_drd,
    mbe=> a_be, msel=> a_sel, mrnw=> a_rnw,
    mseqAddr=> a_seqAddr, mack=> a_ack);  
  
  
end opb_master_arch;